#!/usr/bin/perl
#
# FloodWorld v1.0.a3 (Main Source Code File)
#
# (C)Copyright 2001
# Brian M. King
# All Rights Reserved
#
# This program is provided as is, without warranty. The author is not liable
# for any and all actions as a result of the code herein. 
#
# This program was designed and developed with the intentions of detecting and
# removing "flooders" from an IRC Network. 
#
# Special thanks to PreZ (Creator and author of Magick IRC Services)
# for helping me with his regex god like knowledge.
# 
# Also, special thanks to Shrapnal for helping me with beta
# testing and debugging the software as well as providing   
# the criticism and ideas that helps to make this the best
# software there is =)
#
use IO::Socket;
use lib "lib";
require "config.pm";
require "protocol.pm";
require "a3.pm";

###############
# Check to see if we are going to 
# run any type of debug modes. 
#
# Please check the Debug Options
# in the README for further
# explaination.
$avCount = 0;
if (lc("@ARGV") =~ /-debug/) {
   foreach $av (@ARGV) {
      if (lc($av) eq "-debug") {
         $avCount++;
         chomp($ARGV[$avCount]);
         if ($ARGV[$avCount] =~ /\D/) {
            print "INVALID DEBUG MODE\n";
            print "Please view the Debug Options in the README\n";
            exit;
         }
         if (($ARGV[$avCount] < 1) || ($ARGV[$avCount] > 3)) {
            print "INVALID DEBUG MODE\n";
            print "Please view the Debug Options in the README\n";
            exit;
         }
         $DebugMode = $ARGV[$avCount];
         print "DEBUG MODE set to $DebugMode\n";
         if ($DebugMode == 1) {
            print "DEBUG: Logging all incoming data to debug_" . &getdigitdate .".log\n";
         } elsif ($DebugMode == 2) {
            print "DEBUG: Logging all outgoing data to debug_" . &getdigitdate .".log\n";
         } elsif ($DebugMode == 3) {
            print "DEBUG: Logging all incoming and outgoing data to debug_" . &getdigitdate .".log\n";
         }
         last;
      }
      $avCount++;
   }
}

###############
# Basically a read only mode, will not
# gline/akill any users on the network
# for any reason. Good for beta testing
if (lc("@ARGV") =~ /-noban/) {
   print "NO BAN MODE\n";
   $NoBan = 1;
}

$sock = new IO::Socket::INET (
                                PeerAddr => $RemoteServer,
                                PeerPort => $RemotePort,
                                Proto    => "tcp"
                             );

die "ERROR: Could not open socket to $RemoteServer on $RemotePort" unless $sock;
print "Successfully connected to remote host.\n";

###############
# Call on the sub routine
# that will load all the databases
# that this service uses. This also
# loads the configuration file and
# the help file.
&load_all;
print "Successfully loaded all databases.\n";

if (lc($config{'EncryptPass'}) eq "yes") {
   if (!(-e "md5/fwmd5")) {
      print "ERROR: You have chosed to use encrypted passwords\n";
      print "       but have not yet compiled MD5. Please go into\n";
      print "       the md5 sud directory and type: make\n";
      exit;
   }
}
$LoadTime = time(); # For getting sync time and uptime

$TFGlines = 0; # Count for total glines since restart
$TotalNicks = 0; # Current nicks stored to memory

$ResetGlineSeconds = 600; # Remove GLines from mem after this amount of seconds.
$DoResetGlines = (time() + $ResetGlineSeconds); # First interval

###############
# This is to make sure that no matter
# what there is always a root admin. 
#
# This particular Root Admin is retrieved
# from the fworld.cfg file.
$ops{lc($config{'RootAdmin'})}->{Level} = 400;
$ops{lc($config{'RootAdmin'})}->{Pass} = $config{'RootAdminPass'};

###############
# Ok, now we need to know what protocol
# we are using and exactly how we are
# going to handle incoming data from
# the remote HUB server. This sub 
# routine is defined in the 
# lib/protocol.pm Perl Module. Read
# ALL comments in that file before 
# attempting to add or edit any 
# protocols therein.
&set_protocol_standards;

###############
# The following sends all the needed
# data to the remote HUB server so 
# that we can link to the server.
#
# Each of these sub routines are 
# defined in the lib/protocol.pm
# Perl Module. Make sure that you
# read ALL commands in that file
# before you attempt to add or
# change anything. One wrong thing
# and the entire service is messed
# up.
&SendLinkPass($sock);
print $sock $protocol{'Proto'} . "\n" if ($protocol{'NeedProtoString'});
&SendServerString($sock);
print $sock ":$MyServerName ISON :" . $OperService . "\n";
&SignOnUser($sock,1);
&SignOnUser($sock,2);

$FinishedBurst = 0;

###############
# Are we going to be using the
# debugging settings?
if ($UseDebug) {
   &SendJoinChan($sock,$DebugChannel,$MyClientName);
}

###############
# Does the remote HUB server need any
# special message to signify that we
# are dont sending them our burst
# data?
print $sock $protocol{'EndOfBurst'} . "\n" if ($protocol{'EndOfBurst'});

print "Successfully sent C/N data to remote host.\n";

###############
# Now that we have connected to the
# remote server lets get the ghost
# client into all the channels 
# from the database.
&client_join_chans;

if (!(lc("@ARGV") =~ /-fg/)) {
   &mkpid; # Make this program drop to the background
} else {
   print "FOREGROUND MODE\n";
   $FG = 1;
}

&logit("Service successfully loaded");

$DebugCount = 1; # Failsafe (Lets not flood globops)

if ($DebugMode) {
   open DEBUG, ">logs/debug_" . &getdigitdate . ".log";
}
while ($sock) {
   $temp = <$sock> or &MyExit;
   chomp($temp);
   chop($temp) if ($protocol{'NeedChop'});

   print "$temp\n" if ($FG);
   &logdebug("<-- $temp") if (($DebugMode == 1) || ($DebugMode == 3));

   $SIG{'HUP'} = 'srehash'; # Received SIGHUP from OS -- Call on srehash sub routine
   $SIG{'SEGV'} = 'segfault'; # A programmers worst nightmare!

   ###############
   # This sections of the loops makes sure
   # that all mIRC Color Codes, Bold Char,
   # Underline Char, Reverse Char, and 
   # Regular Char is stripped from any 
   # messages. (High ASCII Char's set in config.pm)
   $temp =~ s/[$Bold,$UnderLine,$Reverse,$Regular]//g;
   $temp =~ s/$Color[0-9]{1,2},[0-9]{1,2}//g;
   $temp =~ s/$Color[0-9]{1,2}//g;
   $temp =~ s/$Color//g;
   $temp =~ s/\?\?\?\?//g;

   $LastMsg = $temp;

   ###############
   # Ok, I was talked into doing this by the all
   # mighty and great programmer PreZ =)
   #
   # This is to make life easier for protocol 
   # determination. Basically if the string 
   # starts with : then we're going to set
   # the scalar $Source with the first parameter
   # because obviously that is the source. Then
   # the $Command will hold the (obviously) command
   # such as MODE, SJOIN, PRIVMSG, NOTICE, etc.
   # If the string did not start with a colon then
   # we know that the first parameter is a command
   # (Same as above.. SJOIN, PRIVMSG, etc) and so
   # lets seperate it all appropriately.
   undef($Source);

   if ($temp =~ /^:/) {
      ($Source,$Command,@RemData) = split(/\s+/,$temp);
      $Source =~ s/://;
      $RemData[0] =~ s/^:// if ($RemData[0] =~ /^:/);
      $RemData[1] =~ s/^:// if ($RemData[1] =~ /^:/);
   } else {
      ($Command,@RemData) = split(/\s+/,$temp);
      $RemData[0] =~ s/^:// if ($RemData[0] =~ /^:/);
      $RemData[1] =~ s/^:// if ($RemData[1] =~ /^:/);
   }

   if ($Command eq "SERVER") {
      $TmpServers = join(" ", @Servers);
      if (!($TmpServers =~ /$RemData[0]/)) {
         push(@Servers, $RemData[0]);
      }
   }
   if ($Command eq "SQUIT") {
      $SQCounter = 0;
      foreach $deadlink (@Servers) {
         if ($deadlink eq $RemData[0]) {
            splice(@Servers,$SQCounter,1);
         } 
         $SQCounter++;
      }
   }
   if (time() >= $DoResetGlines) {
      &remSet;
      $DoResetGlines = (time() + $ResetGlineSeconds);
   }

   if ($Command eq "303") {
      $FinishedBurst = 1;
      $Now = time();
      $Seconds = (time() - $LoadTime);
      &globops("Sync complete -- Start: $LoadTime -- Now: $Now -- $Seconds Second(s)");
      if ($NoBan) {
         &globops("Service running in NO BAN mode");
      }
   }

   ###############
   # The following just makes sure 
   # that we dont ping timeout. 
   if ($Command eq "PING") {
      print $sock "PONG :$MyServerName\n";
   }

   ###############
   # Auto-Identify to Nick Service system.
   #
   # This is purely based on matching the Nick,
   # the Host, and matching the Key Word so that
   # this service knows that it needs to identify.
   #
   # I do say right now that you need to choose 
   # your keyword wisely as if you add something 
   # that the nick services replies with then you
   # can end up causing a major loop which will
   # more than likely crash this software.
   if (($Command eq "NOTICE") && ($Source) && (!($Source =~ /\./))) {
      if ($Source eq $NickServiceNick) {
         if ($nicks{lc($Source)}->{Host} eq $NickServiceHost) {
            if ($UseNickService) {
               if ("@RemData" =~ /$NickServiceKeyWord/) {
                  print $sock ":$MyClientName PRIVMSG $NickServiceNick :IDENTIFY $MyNickPass\n";
                  &logdebug("--> :$MyClientName PRIVMSG $NickServiceNick :IDENTIFY $MyNickPass") if (($DebugMode == 2) || ($DebugMode == 3));
               }
            }
         }
      } elsif ($NoticeFloodStatus eq "on") {
         ###############
         # Notice flood detection
         $nfWhere = $RemData[0];
         splice(@RemData,0,1);
         $nfText = join(" ", @RemData);
         $nfText =~ s/^://;
         if (!(defined($noticeflood{$nfText}))) {
            $noticeflood{$nfText}->{Count} = 1;
            $noticeflood{$nfText}->{Time} = (time() + $NoticeFloodTime);
            $noticeflood{$nfText}->{LastChan} = lc($nfWhere);
         } else {
            if ($noticeflood{$nfText}->{LastChan} eq lc($nfWhere)) {
               $noticeflood{$nfText}->{Count}++;
            }
         }
         $noticeflood{$nfText}->{Count}++;
         if ($noticeflood{$nfText}->{Count} >= $NoticeFloodCount) {
            if ((time() <= $noticeflood{$nfText}->{Time}) && (!(defined($glined{$nicks{lc($Source)}->{Host}})))) {
               if (($nicks{lc($Source)}->{Oper} == 0) && (!(defined($ops{lc($Source)})))) {
                  &globops("Auto-Notice Flood detected -- (In: $nfWhere) -- Text: $nfText");
                  print $sock ":$MyClientName PRIVMSG $DebugChannel :NOTICE FLOOD -- In: $nfWhere -- $Source -- " . $nicks{lc($Source)}->{Host} . " -- $nfText\n" if ($UseDebug);
                  &loggline("NOTICE FLOOD -- $nfWhere -- $Source -- " . $nicks{lc($Source)}->{Host});
                  $glined{$nicks{lc($Source)}->{Host}} = (time() + 600);
                  &RemoveUser($sock,"*\@" . $nicks{lc($Source)}->{Host},"(FloodWorld) Notice Flood Detected -- If you believe this is an error contact $HelpHost immediately");
                  $TFGlines++;
               }
            } else {
               delete($noticeflood{$nfText});
            }
         }
      }
   }

   if (($Command eq "KICK") && ($Source)) {
      if (time() > $nicks{lc($RemData[1])}->{LastKick}) {
         $nicks{lc($RemData[1])}->{LastKick} = (time() + 30);
         $nicks{lc($RemData[1])}->{KickChan} = lc($RemData[0]);
      }
   }
   ###############
   # Ok, this is triggered when (Per the protocol settings) a user
   # joins a channel. It first makes sure that the join flood prot
   # is active, then it makes sure that it is happening in the same
   # channel. We dont want to accidently ban someone if they just 
   # joined all the same channels they normally sit in just because
   # they do it fast.
   if ((($FinishedBurst) && ($JoinFloodStatus eq "on") && ($Command eq $protocol{'JoinTrigger'}))) {
      if (lc($protocol{'JoinNick'}) ne "source") {
         $Source = $RemData[$protocol{'JoinNick'}];
      }
      $Source =~ s/://;
      $Source =~ s/\@//;
      $Source =~ s/\+//;
      $Source =~ s/\%//;
      if (!($joinflood{lc($Source)}->{Time})) {
         $joinflood{lc($Source)}->{Time} = (time() + $JoinFloodTime);
      }
      if (!($joinflood{lc($Source)}->{LastChan})) {
         $joinflood{lc($Source)}->{LastChan} = lc($RemData[$protocol{'JoinChannel'}]);
      }
      if ($joinflood{lc($Source)}->{LastChan} eq lc($RemData[$protocol{'JoinChannel'}])) {
         $joinflood{lc($Source)}->{Count}++;
         if ($joinflood{lc($Source)}->{Count} >= $JoinFloodCount) {
            if ((time() <= $joinflood{lc($Source)}->{Time}) && ($nick{lc($Source)}->{KickChan} ne $joinflood{lc($Source)}->{LastChan})) {
               if (($nicks{lc($Source)}->{Oper} == 0) && (!(defined($ops{lc($Source)})))) {
                  if (!(defined($glined{$nicks{lc($Source)}->{Host}}))) {
                     if (($nicks{lc($Source)}->{Host} ne "") && (!($nicks{lc($Source)}->{LastKick}))) {
                        $glined{$nicks{lc($Source)}->{Host}} = 1;
                        print $sock ":$MyClientName PRIVMSG $DebugChannel :JF (In: " . $joinflood{lc($Source)}->{LastChan} . " -- " . $joinflood{lc($Source)}->{Count} . ") -- $Source -- " . $nicks{lc($Source)}->{Host} . "\n";
                        &logdebug("--> :$MyClientName PRIVMSG $DebugChannel :JF (In: " . $joinflood{lc($Source)}->{LastChan} . " -- " . $joinflood{lc($Source)}->{Count} . ") -- $Source -- " . $nicks{lc($Source)}->{Host}) if (($DebugMode == 2) || ($DebugMode == 3));
                        &loggline("JOIN FLOOD -- $Source -- " . $nicks{lc($Source)}->{Host});
                        &RemoveUser($sock,"*\@" . $nicks{lc($Source)}->{Host},"(FloodWorld) Join Flooding -- If you believe this is an error please contact $HelpHost immediately");
                        &KillUser($sock,$Source,"(FloodWorld) Join flooding") if ($UseKill);
                        $TFGlines++;
                     }
                  }
               }
            } else {
               $joinflood{lc($Source)}->{Count} = 1;
               $joinflood{lc($Source)}->{Time} = (time() + $JoinFloodTime);
            }
         }
      }
   }
   ###############
   # Yes, another loop. This goes through and makes sure
   # we arent wasting any memory by leaving users stored
   # to mem that joined a channel awhile ago but did not
   # do any flooding. The bad part is this will make the
   # service a little slower (About 0.001 seconds but it
   # does add up)
   foreach $CheckRemoveJF (keys %joinflood) {
      if (time() > $joinflood{$CheckRemoveJF}->{Time}) {
         delete($joinflood{$CheckRemoveJF});
      }
   }

   ###############
   # Now the fun begin's =)
   #
   if (($Command eq "PRIVMSG") && (!($Source =~ /\./))) {
      undef(@params);
      $RemData[1] =~ s/://;

      ###############
      # Ok, Why this? Simple actually. The reason for
      # this is that different IRCd's have different
      # formats for sending their PRIVMSG data through
      # in different formats. 
      # IE: Some IRCd's send it as:
      # :Nick PRIVMSG
      # While others use:
      # :Nick!Ident@Host PRIVMSG
      # Yes others still use:
      # :Nick@host
      #
      # So essentially what we are doing is making sure
      # no matter what we are only getting the nickname because, 
      # well, quite frankly all we need is the nickname :) We already
      # have their host stored to memory from the NICK strings passed
      # to us from the remote host server.
      $Source =~ s/[!\@].*//g if ($Source);
      $Source =~ s/[\[,\]]//g;

      $FloodNick = $Source;

      if ($RemData[0] =~ /^\#/) { # Make sure this came from a channel.
         $FloodChan = $RemData[0];  
         ###############
         # This code uses a Hash to keep track of who did
         # the CTCP and how many times did they do it and
         # was it within the specified time frame set by 
         # the CTCPFLOOD command.
         if ($RemData[1] =~ /$CTCPChar/) { # Was this a channel wide CTCP ?
            foreach $CTCPType (@CTCPTypes) {
               $CheckCTCP = $CTCPChar . $CTCPType;
               ##############
               # Due to the fact that for some odd reason
               # actions use the same high ascii character
               # that CTCPs use I had to make it detect it
               # not just by char but also by type.. IE: 
               # PING, VERSION, etc.. So lets do that right
               # here.
               if ($RemData[1] =~ /^$CheckCTCP/) { 
                  if ($CTCPFlood ne "off") {
                     if ($nicks{lc($Source)}->{CTCPTime} == 0) {
                        $nicks{lc($Source)}->{CTCPTime} = (time() + $CTCPFloodTime);
                        $nicks{lc($Source)}->{CTCPs} = 0;
                     } else {
                        $nicks{lc($Source)}->{CTCPs}++;
                        if ($nicks{lc($Source)}->{CTCPs} >= $CTCPFlood) {
                           if (time() <= $nicks{lc($Source)}->{CTCPTime}) {
                              if ($nicks{lc($Source)}->{CTCPTime} != 0) {
                                 if (($nicks{lc($Source)}->{Oper} == 0) && (!(defined($ops{lc($Source)})))) {
                                    print $sock ":$MyClientName PRIVMSG $DebugChannel :CTCPF: $Source -- " . $nicks{lc($Source)}->{Host} . "\n" if ($UseDebug);
                                    &logdebug("--> :$MyClientName PRIVMSG $DebugChannel :CTCPF: $Source -- " . $nicks{lc($Source)}->{Host}) if (($DebugMode == 2) || ($DebugMode == 3));
                                    &loggline("GLINED - CTCP Flood - $Source (" . $nicks{lc($Source)}->{Host} . ")");
                                    &RemoveUser($sock,"*\@" . $nicks{lc($Source)}->{Host},"($MyClientName) Flooding -- If you believe this is an error please contact $HelpHost immediately.");
                                    &KillUser($sock,$Source,"(FloodWorld) CTCP Flooding") if ($UseKill);
                                    $nicks{lc($Source)}->{CTCPTime} = 0;
                                    $nicks{lc($Source)}->{CTCPs} = 0;
                                    $TFGlines++;
                                 }
                              }
                           } else {
                              $nicks{lc($Source)}->{CTCPTime} = 0;
                              $nicks{lc($Source)}->{CTCPs} = 0;
                           }                    
                        }
                     }
                  }
               }
            }
         } else {
            splice(@RemData,0,1);
            $CheckText = join(" ", @RemData);
            $CheckText =~ s/://;

            ###############
            # See now this is really the fun part. 
            # What this is doing is first making sure
            # the text typed in a channel is more than
            # 15 charecters long. If so then lets create
            # a hash entry that carries that text. Then 
            # make sure that it was said in the same channel
            # as previously so that we know its a flood. Now,
            # if its said more than X amount of times in X
            # amount of seconds then assume it flood text and
            # take appropriate action. When assumed flood text
            # it will remain in memory as floodtext for another
            # X amount of seconds.
            if ($misc{'AutoFlood'} ne "off") {
               if (length($CheckText) >= 4) {
                  if (!(defined($FText{$CheckText}->{LastChan}))) {
                     $FText{$CheckText}->{LastChan} = $FloodChan;
                  }
                  if ($FloodChan eq $FText{$CheckText}->{LastChan}) {
                     $FText{$CheckText}->{ChanCount}++;
                     $FText{$CheckText}->{Count}++;
                  }
                  if ($FText{$CheckText}->{Count} == 1) {
                     $FText{$CheckText}->{Time} = (time() + $AutoFloodTime); # X amount of seconds
                  }
                  if ((time() <= $FText{$CheckText}->{Time}) && ($FText{$CheckText}->{IsFlood} != 1)) {
                     if ($FText{$CheckText}->{Count} == $AutoFloodCount) { # X amount of times
                        if (($nicks{lc($FloodNick)}->{Oper} == 0) && (!(defined($ops{lc($FloodNick)})))) { 
                           $FText{$CheckText}->{IsFlood} = 1;
                           $FText{$CheckText}->{Time} = (time() + 180); # Another X amount of seconds
                           &globops("Text ($CheckText) (In: " . $FText{$CheckText}->{LastChan} . ") (Last Said by: $FloodNick) has been said $AutoFloodCount times in less than $AutoFloodTime seconds -- Assuming it Flood Text -- Type (/msg $MyClientName deltext $CheckText) to remove it from memory");
                           $glined{$nicks{lc($FloodNick)}->{Host}} = 1;
                           &loggline("AUTO TEXT FLOOD -- $FloodNick -- " . $nicks{lc($FloodNick)}->{Host} . " -- $CheckText");                            
                           print $sock ":$MyClientName PRIVMSG $DebugChannel :TF (A): $FloodNick -- " . $nicks{lc($FloodNick)}->{Host} . " -- TEXT:($CheckText)\n" if ($UseDebug);
                           &logdebug("--> :$MyClientName PRIVMSG $DebugChannel :TF (A): $FloodNick -- " . $nicks{lc($FloodNick)}->{Host} . " -- TEXT:($CheckText)") if (($DebugMode == 2) || ($DebugMode == 3)); &RemoveUser($sock,"*\@" . $nicks{lc($FloodNick)}->{Host},"(FloodWorld) Flooding -- If you believe this is an error please contact $HelpHost immediately.");                          
                           &KillUser($sock,$FloodNick,"(FloodWorld) Auto Flood Detection") if ($UseKill);
                           $TFGlines++;
                        }
                     }
                  }
                  if ($FText{$CheckText}->{IsFlood} == 1) {
                     if (($nicks{lc($FloodNick)}->{Oper} eq 0) && (!(defined($ops{lc($FloodNick)})))) {
                        if (time() <= $FText{$CheckText}->{Time}) {
                           if (!(defined($glined{$nicks{lc($FloodNick)}->{Host}}))) {
                              $glined{$nicks{lc($FloodNick)}->{Host}} = 1;                            
                              &loggline("AUTO TEXT FLOOD -- $FloodNick -- " . $nicks{lc($FloodNick)}->{Host} . " -- $CheckText");                            
                              print $sock ":$MyClientName PRIVMSG $DebugChannel :TF (A): $FloodNick -- " . $nicks{lc($FloodNick)}->{Host} . " -- TEXT:($CheckText)\n" if ($UseDebug);
                              &logdebug("--> :$MyClientName PRIVMSG $DebugChannel :TF (A): $FloodNick -- " . $nicks{lc($FloodNick)}->{Host} . " -- TEXT:($CheckText)") if (($DebugMode == 2) || ($DebugMode == 3));
                              &RemoveUser($sock,"*\@" . $nicks{lc($FloodNick)}->{Host},"(FloodWorld) Flooding -- If you believe this is an error please contact $HelpHost immediately.");
                              &KillUser($sock,$FloodNick,"(FloodWorld) Auto Flood Detection") if ($UseKill);
                              $TFGlines++;
                           }
                        }
                     }
                  }
               }
            }
            ###############
            # This section is used to make sure that 
            # the text said in a channel is said in a
            # supported channel. Then it goes through
            # and parses the text stored in the database
            # one at a time and compares it so the text
            # said by a user. If it is a match text and 
            # if it meets the X amount of times in X
            # amount of seconds (Set by the TEXTFLOOD command).
            # Note that this matches by key word/phrase said
            # within the text, it does not look for exact full
            # match of the entire line of text from the user.
            foreach $CheckChan (@Channels) {
               chomp($CheckChan);
               if (lc($FloodChan) eq lc($CheckChan)) {
                  if ($ChanFloodTextHandle ne "off") {
                     foreach $CheckFloodText (@ChanFloodText) {
                        @CFT = split(/::/, $CheckFloodText);
                        chomp($CFT[2]);
                        if (($CheckText =~ /$CFT[2]/) && ($CheckFloodText ne "")) {
                           if ($nicks{lc($FloodNick)}->{FTTime} eq 0) {
                              $nicks{lc($FloodNick)}->{FTTime} = (time() + $CFT[1]);
                           }
                           $nicks{lc($FloodNick)}->{FTCount}++;
                           if ($nicks{lc($FloodNick)}->{FTCount} >= $CFT[0]) {
                              if (time() <= $nicks{lc($FloodNick)}->{FTTime}) {
                                 if (($nicks{lc($FloodNick)}->{Oper} eq 0) && (!(defined($ops{lc($FloodNick)})))) {
                                    if (!(defined($glined{$nicks{lc($FloodNick)}->{Host}}))) {
                                       $TFGlines++;
                                       print $sock ":$MyClientName PRIVMSG $DebugChannel :TF (M): $FloodNick -- " . $nicks{lc($FloodNick)}->{Host} . " -- TEXT:($CFT[2]) ($CFT[0] lines in $CFT[1] seconds)\n" if ($UseDebug);
                                       &logdebug(":$MyClientName PRIVMSG $DebugChannel :TF (M): $FloodNick -- " . $nicks{lc($FloodNick)}->{Host} . " -- TEXT:($CFT[2]) ($CFT[0] lines in $CFT[1] seconds)") if (($DebugMode == 2) || ($DebugMode == 3));
                                       $glined{$nicks{lc($FloodNick)}->{Host}} = (time() + $ResetGlineSeconds);
                                       &loggline("GLINE - Text Flood - $FloodNick (" . $nicks{lc($FloodNick)}->{Host} .") -- $CheckFloodText");
                                       &RemoveUser($sock,"*\@" . $nicks{lc($FloodNick)}->{Host},"($MyClientName) Flooding -- If you believe this is an error please contact $HelpHost immediately.");
                                       &KillUser($sock,$FloodNick,"(FloodWorld) Text Flood") if ($UseKill);
                                    }
                                 }
                              } else {
                                 $nicks{lc($FloodNick)}->{FTTime} = (time() + $CFT[1]);
                                 $nicks{lc($FloodNick)}->{FTCount} = 1;
                              }
                           }
                        }
                     }
                  }
               }
            }
         }
      } else {
         ###############
         # Command identification section. What this
         # does is checks to see if this was a message sent 
         # to the service. Then it takes their first word
         # and searches through all commands in the %commands
         # hash (Defined in config.pm) to find a match. This
         # is used so people can do a partial command name
         # or full command name. What this does is upon
         # matching a command it takes all the parameters
         # sent by the user and passes that along to the
         # specified sub routine for the command (Also specified
         # in the commands hash).
         if (lc($RemData[0]) eq lc($MyClientName)) {
            $CMatches = 0;
            $RemData[1] =~ s/://;
            $ComNick = $Source;
            $newCom = lc($RemData[1]);
            splice(@RemData,0,2);
            $parameters = join(" ", $ComNick, @RemData);
            foreach $CheckCom (keys %commands) {
               if ($CheckCom =~ /^$newCom/) {
                  $CMatches++;
                  $DoCom = $CheckCom;
               }
            }
            if ($CMatches == 0) {
               &notice($ComNick,"Command not found");
            } elsif ($CMatches > 1) {
               &notice($ComNick,"More than one command matches $newCom");
            } else {
               eval(q|&{$commands{$DoCom}}($parameters)|);
            }
         }
         if (($nicks{lc($Source)}->{Oper} eq 0) && (!(defined($ops{lc($Source)})))) {
            $msgflood{lc($Source)}->{Count}++;
            if (!(defined($msgflood{lc($Source)}->{Timer}))) {
               $msgflood{lc($Source)}->{Timer} = (time() + 60);
            }
            if ($msgflood{lc($Source)}->{Count} >= 3) {
               if (time() <= $msgflood{lc($Source)}->{Timer}) {
                  &loggline("$Source was flood me so I glined them");
                  print $sock ":$MyClientName PRIVMSG $DebugChannel :PRIVMSG FLOOD --  $Source -- " . $nicks{lc($Source)}->{Host} . "\n" if ($UseDebug);
                  &RemoveUser($sock,"*\@" . $nicks{lc($Source)}->{Host},"($MyClientName) Do not flood a network service");
                  &KillUser($sock,$Source,"(FloodWorld) Dont flood a network service") if ($UseKill);
                  delete($msgflood{lc($Source)});
               } else {
                  undef($msgflood{lc($Source)}->{Timer});
                  $msgflood{lc($Source)}->{Count} = 0;
               }
            }
         }
      }
   }
   ###############
   # This is triggered when user modes are
   # set on a person. Mainly to determine
   # if they are an oper or not. If so 
   # make it known so we dont mess with them
   # for any reason.
   if (($Command eq "MODE") && (!($Source =~ /\./))) {
      if (defined($nicks{lc($RemData[0])})) {
         if ($RemData[1] =~ /o/) {
            if ($RemData[1] =~ /\+/) {
               $nicks{lc($RemData[0])}->{Oper} = 1;
            } else {
               $nicks{lc($RemData[0])}->{Oper} = 0;
            }
         }
      }
   }
   ###############
   # Recevied a NICK string from the remote server
   # so lets store that user to memory including 
   # their Nick, Host, Realname, etc. If 'o' is in
   # their usermodes then lets tag them as an oper
   # then lets see if their realname matches any
   # realnames in the grnames database. If so lets
   # remove them from the network.
   if (($Command eq "NICK") && (!($Source))) {
      $nicks{lc($RemData[$protocol{'NickString'}->{NickName}])}->{Ident} = $RemData[$protocol{'NickString'}->{Ident}];
      $nicks{lc($RemData[$protocol{'NickString'}->{NickName}])}->{Host} = $RemData[$protocol{'NickString'}->{Host}];
      $nicks{lc($RemData[$protocol{'NickString'}->{NickName}])}->{FTTime} = 0;
      $nicks{lc($RemData[$protocol{'NickString'}->{NickName}])}->{NFFirst} = 1;
      if ($protocol{'NickString'}->{UModes}) {
         if ($RemData[$protocol{'NickString'}->{UModes}] =~ /o/) {
            $nicks{lc($RemData[$protocol{'NickString'}->{NickName}])}->{Oper} = 1;
         } else {
            $nicks{lc($RemData[$protocol{'NickString'}->{NickName}])}->{Oper} = 0;
         }
      }
      if (lc($RemData[$protocol{'NickString'}->{NickName}]) eq lc($OperService)) {
         $OperServicePresent = 1;
      }
      $TotalNicks++;
      $RNNick = $RemData[$protocol{'NickString'}->{NickName}];
      splice(@RemData,0,$protocol{'NickString'}->{RealName});
      $RemData[$protocol{'NickString'}->{NickName}] =~ s/://;
      $RN = join(" ", @RemData);
      $nicks{lc($RNNick)}->{RealName} = $RN;
      if ($GRNameStatus eq "on") {
         foreach $line (@GNames) {
            chomp($line);
            if ($line ne "") {
               if ($line eq $RN) {
                  if (($nicks{lc($RNNick)}->{Oper} == 0) && (!(defined($ops{lc($RNNick)})))) {
                     print $sock ":$MyClientName PRIVMSG $DebugChannel :GRNAME (CONNECT): $RNNick -- " . $nicks{lc($RNNick)}->{Host} . " -- REALNAME:($line)\n" if ($UseDebug);
                     &logdebug("--> :$MyClientName PRIVMSG $DebugChannel :GRNAME (CONNECT): $RNNick -- " . $nicks{lc($RNNick)}->{Host} . " -- REALNAME:($line)") if (($DebugMode == 2) || ($DebugMode == 3));
                     &loggline("CONNECT GRNAME GLINE - $RNNick - " . $nicks{lc($RNNick)}->{Host} . " -- $RN");
                     &RemoveUser($sock,"*\@" . $nicks{lc($RNNick)}->{Host},"(FloodWorld) You have triggered a network flood protection service. If you believe this is an error contact $HelpHost immediately.");
                     &KillUser($sock,$RNNick,"(FloodWorld) Flood Detection Triggered") if ($UseKill);
                     $TFGlines++;
                  }
               }
            }
         }
      }
   }
   ###############
   # This is triggered when a user changes their 
   # nickname. When they do that lets copy their
   # old profile in memory to the new nickname 
   # and remove the old profile so that we arent
   # wasting any memory.
   if (($Command eq "NICK") && ($Source)) {
      $nicks{lc($RemData[0])}->{Ident}    = $nicks{lc($Source)}->{Ident};
      $nicks{lc($RemData[0])}->{Host}     = $nicks{lc($Source)}->{Host};
      $nicks{lc($RemData[0])}->{RealName} = $nicks{lc($Source)}->{RealName};
      $nicks{lc($RemData[0])}->{Oper}     = $nicks{lc($Source)}->{Oper};
      $nicks{lc($RemData[0])}->{FTTime}   = $nicks{lc($Source)}->{FTTime};
      $nicks{lc($RemData[0])}->{NFFirst}  = $nicks{lc($Source)}->{NFFirst};
      $nicks{lc($RemData[0])}->{NFTime}   = $nicks{lc($Source)}->{NFTime};
      $nicks{lc($RemData[0])}->{NFCount}  = $nicks{lc($Source)}->{NFCount};
      delete($nicks{lc($Source)});
      if (defined($ops{lc($Source)})) {
         $ops{lc($Source)}->{Identified} = 0;
      }
      $Source = $RemData[0];
      ###############
      # This is triggered if the service nick is changed
      # from something like Services Nick Enforcement
      # This assumes there is no lag and so sends the
      # GHOST command to the NickService incase they 
      # loaded a pseudo client to hold onto the nickname
      # then regain the nickname and identify so that it
      # does not happen again.
      if ($UseNickService) {
         if ($Source eq $MyClientName) {
            print $sock ":$MyClientName PRIVMSG $NickServiceNick :GHOST $MyClientName $MyNickPass\n";
            print $sock ":$RemData[1] NICK $MyClientName\n";
            print $sock ":$MyClientName PRIVMSG $NickServiceNick :IDENTIFY $MyNickPass\n";
            &logdebug("--> :$MyClientName PRIVMSG $NickServiceNick :GHOST $MyClientName $MyNickPass") if  (($DebugMode == 2) || ($DebugMode == 3));
            &logdebug("--> :$RemData[1] NICK $MyClientName") if (($DebugMode == 2) || ($DebugMode == 3));
            &logdebug("--> :$MyClientName PRIVMSG $NickServiceNick :IDENTIFY $MyNickPass") if (($DebugMode == 2) || ($DebugMode == 3));
         }
      }
      ###############
      # Now lets see if we're using Nick Flood protection
      # or not. If its active then lets see how many times 
      # they have changed their nick and if they have done
      # it in less time than the permitted amount. If so
      # then lets remove them from the network. We dont 
      # want nick flooders around =)
      if ($NickFloodStatus eq "on") {
         $nicks{lc($RemData[0])}->{NFCount}++;
         if ($nicks{lc($RemData[0])}->{NFFirst} == 1) {
            $nicks{lc($RemData[0])}->{NFFirst} = 0;
            $nicks{lc($RemData[0])}->{NFTime} = (time() + $NickFloodTime);
         } else {
            if (time() <= $nicks{lc($RemData[0])}->{NFTime}) {
               if ($nicks{lc($RemData[0])}->{NFCount} >= $NickFloodCount) {
                  if (($nicks{lc($RemData[0])}->{Oper} == 0) && (!(defined($ops{lc($RemData[0])})))) {
                     print $sock ":$MyClientName PRIVMSG $DebugChannel :NF: $RemData[0] -- " . $nicks{lc($RemData[0])}->{Host} . " -- $NickFloodCount changes in under $NickFloodTime seconds\n" if ($UseDebug);
                     &logdebug("--> :$MyClientName PRIVMSG $DebugChannel :NF: $RemData[0] -- " . $nicks{lc($RemData[0])}->{Host} . " -- $NickFloodCount changes in under $NickFloodTime seconds") if (($DebugMode == 2) || ($DebugMode == 3));
                     &loggline("NICK FLOOD -- $RemData[0] -- " . $nicks{lc($RemData[1])}->{Host});
                     &RemoveUser($sock,"*\@" . $nicks{lc($RemData[0])}->{Host},"(FloodWorld) Nick Flood Detected -- If you believe this is an error please contact $HelpHost immediately.");
                     &KillUser($sock,$RemData[0],"(FloodWorld) Nick Flooding") if ($UseKill);
                  }
               }
            } else {
               $nicks{lc($RemData[0])}->{NFFirst} = 0;
               $nicks{lc($RemData[0])}->{NFCount} = 0;
            }
         }
      }
   }
   ###############
   # Ahh, a user has just logged off the network
   # (For whatever reason) so lets remove them
   # from memory and make sure they are no longer
   # tagged as identified. This makes sure that
   # A) We arent wasting any memory and B) No one
   # can switch to that nick and try and take their
   # place as an oper and be identified.
   if (uc($Command) eq "QUIT") {
      if (lc($Source) eq lc($OperService)) {
         $OperServicePresent = 0;
      }
      delete($nicks{lc($Source)});
      if (defined($ops{lc($Source)})) {
         $ops{lc($Source)}->{Identified} = 0;
      }
      $TotalNicks--;
   }

   ###############
   # This is here in reference to the auto-detect
   # text flood system. What it does is goes through
   # all the text stored to memory and checks how long
   # they have been there. If the current time is 
   # greater than the time set for that text to be
   # in memory then delete it from memory so we dont
   # waste any memory and so we dont accidently assume
   # some text as flood text when it should have been
   # removed in the first place.
   foreach $FTRem (keys %FText) {
      if (time() > $FText{$FTRem}->{Time}) {
         delete($FText{$FTRem});
      }
   }
}
sub MyExit {
   &logit("Lost Connection -- Last Sub: $LastSub -- Last Message: $LastMsg");
   exit;
}
###############
# ACK! We got a seg fault on our hands.
sub segfault {
   &globops("AHHHH! Seg Fault! Last Sub -- $LastSub");
   &logit("SEGFAULT -- Last Sub -- $LastSub");
   exit;
}
###############
# Sub routine to load all the databases and the
# help file. The reason this is all put into a
# single sub routine is to lower the amount of
# routines sitting on the program stack. I
# left all the writing sub routines as is 
# because they are called on at different times
# and only when needed so there is no point in
# doing more processing then we have too.
sub load_all {
   $LastSub = "load_all";
   # Text not allowed to be added with TEXTFLOOD
   if (-e "forbidden.db") {
      open FORB, "forbidden.db";
         chmod 0777, FORB;
         @forbidden = <FORB>;
      close FORB;
      &logit("Successfully loaded forbidden text database");
   }
   # The help file (DUH)
   if (-e "help.txt") {
      open HELP, "help.txt";
         @Help = <HELP>;
      close HELP;
      &logit("Successfully loaded help file");
   } else {
      &logit("Could not locate help.txt");
      die "Could not locate help file.";
   }
   # Configuration File
   if (-e "fworld.cfg") {
      open CONF, "fworld.cfg";
         @TmpConfig = <CONF>;
      close CONF;
      foreach $ConfItem (@TmpConfig) {
         chomp($ConfItem);
         @Item = split(/\=/, $ConfItem);
         $config{$Item[0]} = $Item[1];
      }
      &logit("Successfully loaded configuration file");
   } else {
      &logit("Could not locate fworld.cfg");
      die "Unable to locate fworld.cfg.";
   }
   # Users database
   if (-e $UserDatabase) {
      open UDB, "$UserDatabase";
         @TempUDB = <UDB>;
      close UDB;
      $UDBCount = 0;
      foreach $line (@TempUDB) {
         chomp($line);
         @TempData = split(/::/, $line);
         if (@TempData > 3) {
            print "ERROR: Too many parameters in a string from the users database.\n";
            print "       I am going to assume you have not yet converted your own\n";
            print "       users database. Please read the README on Converting your\n";
            print "       Users Database.\n";
            exit;
         }
         if (@TempData < 3) {
            if ($TempData[0] ne "") {
               &logit("ERROR: There is an error in $UserDatabase");
               print "ERROR: There is an error in $UserDatabase\n";
               print "Attempting to compensate.\n";
               splice(@TempUDB, $UDBCount, 1);
               next;
            }
         }
         $ops{$TempData[0]}->{Pass} = $TempData[1];
         $ops{$TempData[0]}->{Level} = $TempData[2];
         $UDBCount++;
      }
      &logit("Successfully loaded users database");
   }
   # Channels Database
   if (-e $ChannelsDatabase) {
      open CHANS, "$ChannelsDatabase";
         @Channels = <CHANS>;
      close CHANS;
      foreach $line (@Channels) {
         chomp($line);
         if ((!($line =~ /\#/)) && ($line ne "")) {
            &logit("ERROR: There is an error in $ChannelsDatabase");
            die "ERROR: There is an error in $ChannelsDatabase";
         }
      }
      &logit("Successfully loaded channels database");
   }
   # TEXTFLOOD database
   if (-e "flood.db") {
      open FD, "flood.db";
         @ChanFloodText = <FD>;
      close FD;
      &logit("Successfully loaded textflood database");
   }
   # GRNames Database
   if (-e "grnames.db") {
      open DB, "grnames.db";
         @GNames = <DB>;
      close DB;
      &logit("Successfully loaded grnames database");
   }
   # Misc. Database
   if (-e "misc.db") {
      open MDB, "misc.db";
         @MDBData = <MDB>;
      close MDB;
      foreach $line (@MDBData) {
         chomp($line);
         @TempMDB = split(/\=/, $line);
         $misc{$TempMDB[0]} = $TempMDB[1];
      }
   }
   &set_flood_settings;
}
###############
# This sub routine is used to determine
# a text returned value of the readable 
# time frame. You pass this sub routine
# any amount of seconds and it will 
# automatically determine how much time
# that is in a readable formate.
#
# IE: &duration(122);
# Returns: 2 minute(s) 2 second(s)
sub duration {
   $LastSub = "duration";
   undef($retval);
   $weeks = int($_[0] / 604800) . " week(s) ";
   $days = int(($_[0] / 86400) % 7) . " day(s) ";
   $hours = int(($_[0] / 3600) % 24) . " hour(s) ";
   $mins = int(($_[0] / 60) % 60) . " minute(s) ";
   $seconds = int($_[0] % 60) . " second(s) ";
   foreach $Add ($weeks,$days,$hours,$mins,$seconds) {
      @tmpTime = split(/\s+/, $Add);
      if ($tmpTime[0] > 0) {
         $retval = $retval . $Add;
      }
   }
   undef(@tmpTime);
   return $retval;
}
###############
# This is obviously used to send a notice
# to the specified person. Its small but
# trust me its worth it. 
#
# IE: &notice("Raider","This is a notice");
# Will send me a notice from FloodWorld 
sub notice {
   print $sock ":$MyClientName NOTICE $_[0] :$_[1]\n";
   &logdebug("--> :$MyClientName NOTICE $_[0] :$_[1]") if (($DebugMode == 2) || ($DebugMode == 3));
}
###############
# As with the above sub routine, this makes it 
# easier to send a globops message. Instead of
# having to write out print and $sock and all
# that other stuff you just use this sub 
# routine.
#
# IE: &globops("Message here");
sub globops {
   print $sock ":$MyClientName GLOBOPS :$_[0]\n" if ($protocol{'HasGlobOps'});
   &logdebug("--> :$MyClientName GLOBOPS :$_[0]") if (($DebugMode == 2) || ($DebugMode == 3));
}
###############
# This is called from the main loop to
# make sure that we remove glined from
# memory. Basically everytime a gline is
# set it is put into memory to make sure
# that we dont try and ban the same address 
# more than once within a certain period 
# of time (Prevents flooding).
sub remSet {
   $LastSub = "remSet";
   foreach $NeedDelete (keys %glined) {
      if (time() >= $glined{$NeedDelete}) {
         delete($glined{$NeedDelete});
      }
   }
   return;
}
###############
# Well, lets make life easier for
# you. Instead of you needing to
# include an ampersand at the command
# line this sub routine is called to 
# drop the program to the background 
# for you. And yes, it even creates
# the pid file.
sub mkpid {
   $tmp = fork;
  
   if ($tmp) {
      open PIDFILE, ">$PIDFile";
         chmod 0777, PIDFILE;   
         print PIDFILE "$tmp\n";
      close PIDFILE;
      exit;
   }
}
###############
# Mainly used for the logging system. 
# What this sub routine does is it
# takes the current date from localtime()
# and converts it to an MMDDYYYY format.
# IE: 02022001 is February 2nd 2001
sub getdigitdate {
   undef($retval);
   ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);

   $mon++; 
   $mon = "0" . $mon if ($mon < 10);
   $year += 1900;
   $mday = "0" . $mday if ($mday < 10);
   $retval = $mon . $mday . $year;
   undef($year);
   return $retval;
}  
###############
# Logging sub routine. This sub will
# read in any existing data from the 
# log file and append to it and write 
# it all to the log file. This sub
# creates logs in name format of
# fworld_MMDDYYYY.log
sub logit {
   $What = shift;
   open LF, "logs/fworld_" . &getdigitdate . ".log";
      @TmpLog = <LF>;
   close LF;
   push @TmpLog, localtime() . ": " . $What . "\n";
   open LF, ">logs/fworld_" . &getdigitdate . ".log";
      chmod 0777, LF;
      foreach $line (@TmpLog) {
         chomp($line);
         if ($line ne "") {
            print LF "$line\n";
         }
      }
   close LF;
   undef(@TmpLog);
}
###############
# Same as above except that it used
# gline_MMDDYYYY.log log file name
# format.
sub loggline {
   $What = shift;
   open GF, "logs/gline_" . &getdigitdate . ".log";
      @TmpLog = <GF>;
   close GF;
   push @TmpLog, localtime() . ": " . $What . "\n";
   open GF, ">logs/gline_" . &getdigitdate . ".log";
      chmod 0777, GF;
      foreach $line (@TmpLog) {
         chomp($line);
         if ($line ne "") {
            print GF "$line\n";
         }
      }
   close GF;
   undef(@TmpLog);
}
###############
# Debug Logging sub routine. This sub will
# read in any existing data from the 
# log file and append to it and write 
# it all to the log file. This sub
# creates logs in name format of
# debug_MMDDYYYY.log
sub logdebug {
   $What = shift;
   print DEBUG "$What\n";
}
################
# Used to match text with wildcard text.
# Main functionality is to see if a user has
# access on a specified channel, if a banned
# host is protected, etc ..
#
# Syntax: &match("Normal Text","Wild Text);
# IE: &match("Hello there","H*o*re"); -- Returns 1 (True);
# IE: &match("adsl-123.test.new.com","adsl*.*.new.com"); -- Returns 1 (True);
# IE: &match("adsl-123.test.new.com","*.test2.new.com"); -- Returns 0 (False);
# IE: &match("adsl-123.test.new.com","adsl-???.????.new.com); -- Returns 1 (True);
# 
# Special thanks to PreZ for helping to create and perfect this and the
# following Sub Routine's.
sub match {
   $LastSub = "match";
   $realhost = shift; 
   $wcstring = shift; 
   $wcstring =~ s/\./\\\./g;
   $wcstring =~ s/\*/\.\*/g;  
   $wcstring =~ s/\?/\./g;
   $wcstring =~ s/^.*\@// unless ($realhost =~ /\@/);
   $wcstring =~ s/^.*\!// unless ($realhost =~ /\!/);
   if ($realhost =~ /^$wcstring$/) { 
      return 1; 
   }
   undef($wcstring);
   return 0;
}
###############
# Take a regular full host as a parameter
# and turn it into a wild card host.
#
# IE: &address("123.bot.com");
# returns: *.bot.com
sub address {
   $LastSub = "address";
   $tmpAdd = shift;
   $nohide = 1;
   $nohide2 = 1;
   if ($tmpAdd =~ /^(\d+\.){3}\d+$/) { 
      $tmpAdd =~ s/(.*)\.([^.]+\.){$nohide2}[^.]+$/$1/g; 
      for($i=0; $i<=$nohide2; $i++) { 
         $tmpAdd = $tmpAdd . ".*"; 
      }; 
   } else {
      $ext = $tmpAdd; 
      $ext =~ s/.*\.(([^.]+\.){$nohide}[^.]+)$/$1/g; 
      $tmpAdd =~ s/[^.]*\d+[^.]*\./\*\./g; 
      $tmpAdd =~ s/([^.]+\.){$nohide}[^.]+$/$ext/g;
   }
   undef($tmpAdd);
   return $tmpAdd;
}
##############
# Generally this sub routine goes
# through and checks first to see
# if the specified person is an 
# operator. If so then lets make
# sure they are an oper. If so
# then lets make sure they are
# identified. If the person passes
# all three of these checks then 
# their access level is returned.
#
# IE: &has_access("Raider");
sub has_access {
   $LastSub = "has_access";
   if (defined($ops{lc($_[0])})) {
      if ($nicks{lc($_[0])}->{Oper} == 1) {
         if ($ops{lc($_[0])}->{Identified}) {
            return $ops{lc($_[0])}->{Level};
         }
      }
   }
   return 0;
}
###############
# Sub routines that write out all
# the databases this service uses.
# Note that these are all written
# as individual sub routines for
# the main purpose of being able
# to be called on without having
# to write out all the databases
# which in turn saves processing 
# time and lowers the chances for
# messing anything up.
sub write_forbidden {
   $LastSub = "write_forbidden";
   open FORB, ">forbidden.db";
      chmod 0777, FRB;
      foreach $WriteForb (@forbidden) {
         chomp($WriteForb);
         if ($WriteForb ne "") {
            print FORB "$WriteForb\n";
         }
      }
   close FORB;
}
sub write_ops {
   $LastSub = "write_ops";
   open UDB, ">$UserDatabase";
      chmod 0777, UDB;
      foreach $WU (keys %ops) {
         print UDB $WU ."::" . $ops{$WU}->{Pass} . "::" . $ops{$WU}->{Level} . "\n";
      }
   close UDB;
}
sub write_chans {
   $LastSub = "write_chans";
   open CHANS, ">$ChannelsDatabase";
      chmod 0777, CHANS;
      foreach $WriteChan (@Channels) {
         chomp($WriteChan);
         if ($WriteChan ne "") {
            print CHANS "$WriteChan\n";
         }
      }
   close CHANS;
}
sub write_flood {
   $LastSub = "write_flood";
   open FD, ">flood.db";
      chmod 0777, FD;
      foreach $WriteFlood (@ChanFloodText) {
         chomp($WriteFlood);
         if ($WriteFlood ne "") {
            print FD "$WriteFlood\n";
         }
      }
   close FD;
}
sub write_grnames {
   $LastSub = "write_grnames";
   open DB, ">grnames.db";
      chmod 0777, DB;
      foreach $line (@GNames) {
         chomp($line);
         if ($line ne "") {
            print DB "$line\n";
         }
      }
   close DB;
}
sub write_misc {
   $LastSub = "write_misc";
   open MDB, ">misc.db";
      chmod 0777, MDB;
      foreach $mWrite (keys %misc) {
         chomp($mWrite);
         if ($misc{$mWrite} ne "") {
            print MDB "$mWrite\=" . $misc{$mWrite} . "\n";
         }
      }
   close MDB;
}
###############
# This routine makes the ghost client
# join all the supported channels
# from the database.
sub client_join_chans {
   $LastSub = "client_join_chans";
   open CHANS, "$ChannelsDatabase";
      @ClientChans = <CHANS>;
   close CHANS;
   foreach $CJoin (@ClientChans) {
      chomp($CJoin);
      if ($CJoin ne "") {
         &SendJoinChan($sock,$CJoin,$GhostNick);
      }
   }
}
###############################################################################
# Remote command sub routines                                                 #
###############################################################################
# One thing you will notice in all the
# command sub routines is that the use
# an array called params that holds all
# the parameters. It is all parsed using
# space as the delimiter.
#
###############
# The remote command DELTEXT which
# is used to delete certain text from
# memory that was set by the auto-detect
# flooding code before it expires on
# its own.
sub do_deltext {
   $LastSub = "do_deltext";
   @params = split(/\s+/, shift);
   $Who = $params[0];
   splice(@params,0,1);
   $What = join(" ", @params);
   if (&has_access($Who) >= 100) {
      if (defined($FText{$What})) {
         delete($FText{$What});
         &globops("$Who removed text ($What) from assumed flood text memory");
      }
   }
   undef(@params);
}
###############
# Remote command LISTADMIN
#
# Displays all the operators on
# the access list and their level
# on the service.
sub do_listadmin {
   $LastSub = "do_listadmin";
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) >= 100) {
      &notice($params[0],"Listing users with access");
      foreach $ListOps (keys %ops) {
         &notice($params[0],"$ListOps -- " . $ops{$ListOps}->{Level});
      }
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed listadmin (No level/ident)");
      &logit("$params[0] failed listadmin");
   }
   undef(@params);
   return;
}
###############
# Remote command ADDOP
#
# Takes parameters that include the
# NickName of the person to be added
# (They must be online) and the level
# in which to add them to access.
sub do_addop {
   $LastSub = "do_addop";
   @params = split(/\s+/, shift);
   if ($params[2] =~ /\D/) {
      &notice($params[0],"Syntax: ADDOP NickName Level");
      return;
   }
   if (@params < 3) {
      &notice($params[0],"Syntax: ADDOP NickName Level");
      return;
   }
   # Lets make sure that it is a root admin (Level 400)
   # doing the command and make sure that we can access
   # the needed data. If so then lets make sure that 
   # the specified level is allowed.
   if (&has_access($params[0]) == 400) {
      if (defined($nicks{lc($params[1])})) {
         if (($params[2] < &has_access($params[0])) || (&has_access($params[0]) == 400)) {
            if ($params[2] <= 400) {
               $ops{lc($params[1])}->{Host} = &address($nicks{lc($params[1])}->{Host});
               $ops{lc($params[1])}->{Ident} = $nicks{lc($params[1])}->{Ident};
               $ops{lc($params[1])}->{Ident} =~ s/\~//;
               $ops{lc($params[1])}->{Level} = $params[2];
               $ops{lc($params[1])}->{Pass} = "none";
               &write_ops;
               &notice($params[1],"You have been added to my database at level $params[2] -- Please set a password");
               &globops("$params[0] added $params[1] to access at $params[2]");
               &logit("$params[0] added $params[1] to access at level $params[2]");
            } else {
               &notice($params[0],"Level must be between 100 and 400");
               &globops("$params[0] failed addop (Bad level)");
               &logit("$params[0] failed addop because they used a bad level");
            }
         } else {
            &notice($params[0],"You cannot set someone to an access level higher than yourself");
            &globops("$params[0] failed addop (Attempted to set $params[1] to a level higher than they are)");
            &logit("$params[0] failed addop $params[1] -- they tried to use a level higher than their own");
         }
      } else {
         &notice($params[0],"Unable to access host for $params[1]");
      }
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed addop (No access/ident)");
      &logit("$params[0] failed to addop $params[1] because they did not have access");
   }
   undef(@params);
   return;
}
###############
# Remote command DELOP
#
# Used to remove an operator from
# access. 
sub do_delop {
   $LastSub = "do_delop";
   @params = split(/\s+/, shift);
   if (@params < 2) {
      &notice($params[0],"Syntax: DELOP NickName Level");
      return;
   }
   if (&has_access($params[0]) == 400) {
      if (defined($ops{lc($params[1])})) {
         if (($ops{lc($params[1])} < &has_access($params[0])) || (&has_access($params[0]) == 400)) {
            delete($ops{lc($params[1])});
            &globops("$params[0] removed $params[1] from access");
            &write_ops;
            &logit("$params[0] removed $params[1] from access");
         } else {
            &notice($params[0],"You cannot remove someone with a higher access than you");
            &globops("$params[0] failed DELOP (No level)");
            &logit("$params[0] failed DELOP -- $params[1] has a higher access level");
         }
      } else {
         &notice($params[0],"$params[1] not found in database");
      }
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed DELOP (No access/ident)");
      &logit("$params[0] failed DELOP $params[1] -- did not have access");
   }
   undef(@params);
   return;
}
###############
# Remote command SETPASS
#
# Used to set a password for the person
# that did the command. This is also 
# used to change an existing password
# of the user that did the command.
sub do_setpass {
   $LastSub = "do_setpass";
   @params = split(/\s+/, shift);
   if (@params < 2) {
      &notice($params[0],"Syntax: SETPASS Password");
      return;
   }
   if (defined($ops{lc($params[0])})) {
      if ($nicks{lc($params[0])}->{Oper} == 1) {
         if (lc($config{'EncryptPass'}) eq "yes") {
            $I_TempPass = &encrypt($params[1]);
         } else {
            $I_TempPass = $params[1];
         }
         $ops{lc($params[0])}->{Pass} = $I_TempPass;
         $ops{lc($params[0])}->{Identified} = 1;
         &write_ops;
         &logit("$params[0] successfully set a password");
         &notice($params[0],"Password for $params[0] set to $params[1]");
      } else {
         &notice($params[0],"Access denied (Non-Oper)");
         &globops("$params[0] failed setpass (Non-Oper)");
         &logit("$params[0] failed setpass -- Non-Oper");
      }
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed setpass (No access)");
      &logit("$params[0] failed setpass -- No access");
   }
   undef(@params);
   return;
}
###############
# Remote command IDENTIFY
#
# This is obviously used for the
# person to identify to the service
# so that they can use it.
sub do_identify {
   $LastSub = "do_identify";
   @params = split(/\s+/, shift);
   if (@params < 2) {
      &notice($params[0],"Syntax: IDENTIFY Password");
      return;
   }
   if (defined($ops{lc($params[0])})) {
      if ($nicks{lc($params[0])}->{Oper} == 1) {
         if ($ops{lc($params[0])}->{Pass} eq "none") {
            &notice($params[0],"You do not have a password set");
            return;
         }
         if (lc($config{'EncryptPass'}) eq "yes") {
            $I_TempPass = &encrypt($params[1]);
         } else {
            $I_TempPass = $params[1];
         }
         if ($ops{lc($params[0])}->{Pass} eq $I_TempPass) {  
            $ops{lc($params[0])}->{Identified} = 1;
            &notice($params[0],"You are now identified");
            &logit("$params[0] successfully identified");
            print $sock ":$MyClientName PRIVMSG $DebugChannel :$params[0] successfully identified.\n" if ($UseDebug);
         } else {
            &notice($params[0],"Access denied (Password Mismatch)");
            &globops("$params[0] failed IDENTIFY (Password Mismatch)");
            &logit("$params[0] failed to IDENTIFY -- Password Mismatch");
         }
      } else {
         &globops("$params[0] failed IDENTIFY (Non-Oper)");
      }
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed IDENTIFY (No access)");
      &logit("$params[0] failed IDENTIFY -- No access");
   }
   undef(@params);
   return;
}
###############
# Remote command ADDCHAN
#
# Used to add a channel to the
# database. This makes it so that
# the ghost client will join this
# channel and that the channel will
# be protected from floods.
sub do_addchan {
   $LastSub = "do_addchan";
   @params = split(/\s+/, shift);
   if (@params < 2) {
      &notice($params[0],"Syntax: ADDCHAN #Channel");
      return;
   }
   if (&has_access($params[0]) >= 100) {
      foreach $ChanExists (@Channels) {
         chomp($ChanExists);
         if (lc($ChanExists) eq lc($params[1])) {
            &notice($params[0],"Channel $params[1] already exists");
            return;
         }
      }
      push @Channels, lc($params[1]) . "\n";
      &notice($params[0],"Added channel $params[1] to database");
      &globops("$params[0] added channel $params[1] to database");
      &SendJoinChan($sock,$params[1],$GhostNick);
      &write_chans;
      &logit("$params[0] added channel $params[1] to database");
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed ADDCHAN (No access/ident)");
      &logit("$params[0] failed ADDCHAN -- No access");
   }
   undef(@params);
}
###############
# Remote command DELCHAN
#
# Used to remove a channel from
# the database. Basically the 
# channel will no longer be protected 
# from flooding.
sub do_delchan {
   $LastChan = "do_delchan";
   @params = split(/\s+/, shift);
   if (@params < 2) {
      &notice($params[0],"Syntax: DELCHAN \#Channel");
      return;
   }
   $CCount = 0;
   if (&has_access($params[0]) >= 100) {
      foreach $ChanExists (@Channels) {
         chomp($ChanExists);
         if (lc($ChanExists) eq lc($params[1])) {
            splice(@Channels,$CCount,1);
            &notice($params[0],"Removed channel $params[1] from database");
            &globops("$params[0] removed channel $params[1] from database");
            print $sock ":$GhostNick PART $params[1] :Channel not supported any longer.\n";
            &logdebug("--> :$GhostNick PART $params[1] :Channel not supported any longer.") if (($DebugMode == 2) || ($DebugMode == 3));
            &write_chans;
            &logit("$params[0] removed channel $params[1] from database");
            return;
         }
         $CCount++;
      }
      &notice($params[0],"Channel $params[1] does not exist");
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed DELCHAN (No access/ident)");
      &logit("$params[0] failed DELCHAN -- No access");
   }
   undef(@params);
}
###############
# Remote command LISTCHANS
#
# This command is used to display
# to the person that did the command
# all the channels that are currently 
# beng supported and protected.
sub do_listchans {
   $LastSub = "do_listchans";
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) >= 100) {
      &notice($params[0],"Listing supported channels");
      foreach $ListChan (@Channels) {
         chomp($ListChan);
         if ($ListChan ne "") {
            &notice($params[0],"$ListChan");
         }
      }
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed LISTCHANS (No access/ident)");
      &logit("$params[0] failed LISTCHANS -- No access");
   }
   undef(@params);
}
###############
# Remote command TEXTFLOOD
#
# This command is used to add/remove 
# text from the database. What this 
# means is if you add specified text
# then anyone (Aside from opers) that
# says anything with that text in it
# in any supported channels will be
# removed from the network.
sub do_textflood {
   $LastSub = "do_textflood";
   @params = split(/\s+/, shift);
   if (@params < 2) {
      &notice($params[0],"Syntax: TEXTFLOOD {ADD|DEL|OFF|ON|LIST} NumberOfLines NumberOfSeconds Text to detect");
      return;
   }

   $pNick = $params[0];
   $pWhat = lc($params[1]);
   $pLines = $params[2];
   $pSeconds = $params[3];

   splice(@params,0,4);
   $pText = join(" ", @params);

   # Lets make sure this text isnt forbidden
   foreach $CheckForb (@forbidden) {
      chomp($CheckForb);
      if (lc($pText) eq lc($CheckForb)) {
         &notice($pNick,"The text ($pText) is forbidden");
         &globops("$pNick attempted to add forbidden text ($pText)");
         &logit("$pNick attempted to add forbidden text ($pText)");
         return;
      }
   }

   # Lets make sure the specified text is atleast
   # 5 characters long.
   if ((length($pText) < 5) && ($pWhat eq "add")) {
      if ($pText eq "") {
         &notice($pNick,"Syntax: TEXTFLOOD ADD NumberOfLines NumberOfSeconds Text to detect");
         return;
      }
      &notice($pNick,"Text must be atleast 5 charecters long.");
      &globops("$pNick attempted to add bad text ($pText) (Too short)");
      return;
   }

   if (&has_access($pNick) >= 200) {
      if ($pWhat eq "add") {
         $ChanFloodTextHandle = "on";
         if (($pLines =~ /\D/) || ($pSeconds =~ /\D/)) {
             &notice($pNick,"Syntax: TEXTFLOOD ADD NumberOfLines NumberOfSeconds Text to detect");
             return;
         }
         foreach $DoesExist (@ChanFloodText) {
            chomp($DoesExist);
            @Exist = split(/::/, $DoesExist);
            if ($Exist eq $pText) {
               &notice($pNick,"That text already exists in the database");
               return;
            }
         }
         if ($pSeconds < 1) {
            &notice($pNick,"Invalid amount of seconds. Must be 1 or greater");
            return;
         } 	
         push @ChanFloodText, $pLines . "::" . $pSeconds . "::" . $pText . "\n";
         &write_flood;
         &notice($pNick,"Added text ($pText) to memory at $pLines lines in $pSeconds seconds");
         &globops("$pNick added text ($pText) to flood text memory at $pLines lines in $pSeconds seconds");
         &logit("$pNick added text -- $pText -- to database at $pLines lines in $pSeconds seconds");
      } elsif ($pWhat eq "off") {
         $ChanFloodTextHandle = "off";
         &notice($pNick,"Flood text deactivated");
         &globops("$pNick deactivated flood text");
         &logit("$pNick deactivated flood text detection");
         $misc{'TextFlood'} = "off";
         &write_misc;
      } elsif ($pWhat eq "on") {
         $ChanFloodTextHandle = "on";
         &notice($pNick,"Flood text activated");
         &globops("$pNick activated flood text");
         &logit("$pNick activated flood text detection");
         $misc{'TextFlood'} = "on";
         &write_misc;
      } elsif ($pWhat eq "del") {
         chomp($ChanFloodText[$pLines]);
         if ($pLines =~ /\D/) {
            &notice($pNick,"Syntax: TEXTFLOOD DEL TextNumber");
            return;
         }
         if ($ChanFloodText[$pLines] eq "") {
            &notice($pNick,"Entry $pText does not exist");
            return;
         }
         @DelFT = split(/::/, $ChanFloodText[$pLines]);
         &notice($pNick,"Flood text number $pLines (" . $DelFT[2] . ") removed");
         &globops("$pNick removed flood text number $pLines (" . $DelFT[2] . ")");
         &logit("$pNick removed text -- " . $DelFT[2] . " -- from database");
         splice(@ChanFloodText,$pLines,1);
         &write_flood;
      } elsif ($pWhat eq "list") {
         $FCount = 0;
         &notice($pNick,"Listing all supported flood text");
         foreach $DispFloodText (@ChanFloodText) {
            @DispFT = split(/::/, $DispFloodText);
            &notice($pNick,"Flood Text $FCount ($DispFT[0] lines in $DispFT[1] seconds): $DispFT[2]");
            $FCount++;
         }
      } else {
         &notice($pNick,"Syntax: TEXTFLOOD {ADD|DEL|LIST|ON|OFF} NumberOfLines NumberOfSeconds (TEXT)");
      }
   } else {
      &notice($pNick,"Access denied");
      &globops("$pNick failed TEXTFLOOD (No access/ident)");
      &logit("$pNick failed TEXTFLOOD -- No access");
   }
   undef(@params);
}
###############
# Remote command CTCPFLOOD
#
# When activated this sets the
# ctcp amount of times and number
# of seconds to watch between.
sub do_ctcpflood {
   $LastSub = "do_ctcpflood";
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) >= 400) {
      if (lc($params[1]) eq "off") {
         $CTCPFlood = "off";
         &notice($params[0],"CTCP flood deactivated");
         &globops("$params[0] deactivated CTCP Flood Protection");
         &logit("$params[0] deactivated CTCP Flood Protection");
         $misc{'CTCPFlood'} = "off";
         &write_misc;
         return;
      } elsif (($params[1] =~ /\D/) || ($params[2] =~ /\D/)) {
         &notice($params[1],"Syntax: CTCPFLOOD NumberOfCTCPs NumberOfSeconds");
         return;
         if (@params < 2) {
            &notice($params[0],"Syntax: CTCPFLOOD {OFF} NumberOfCTCPs NumberOfSeconds");
            return;
         }
      } else {
         $CTCPFlood = $params[1];
         $CTCPFloodTime = $params[2];
         &notice($params[0],"CTCP Flood set to $params[1] CTCPs in $params[2] seconds");
         &globops("$params[0] set CTCP Flood to $params[1] CTCPs in $params[2] seconds");
         &logit("$params[0] set ctcp flood to $params[1] CTCPs in $params[2] Seconds");
         $misc{'CTCPFlood'} = $params[1] . "::" . $params[2];
         &write_misc;
      }
   } else {
      &notice($params[0],"Access denied");
      globops("$params[0] failed CTCPFLOOD (No access/ident)");
      &logit("$params[0] failed CTCPFLOOD -- No access");
   }
   undef(@params);
}
###############
# Remote command HELP
#
# Ahh. This has to be one of my
# favorite parts of the entire
# service.
#
# What this does is depending on if
# a command was specified or if the
# user just sent HELP this routine
# will read through the help file
# (Stored to memory) to find the 
# section in which the specified help
# was requested. Please view the help.txt
# file to see how the sections are setup.
sub do_help {
   $LastSub = "do_help";
   @params = split(/\s+/, shift);
   $HCount = 0;
   if (!(&has_access($params[0]))) {
      &notice($params[0],"You have either not yet identified or do not have access to view the help.");
      if ($nicks{lc($params[0])}->{Oper} eq 1) {
         if ($ops{lc($params[0])}->{Pass} eq "none") {
            &notice($params[0],"I have notice you have not yet set a password for your nick name. Please use the setpass command to set a password.");
            &notice($params[0],"Syntax: SETPASS Password");
         }
      }
      return;
   }   
   if ($params[1]) {
      foreach $HelpNumber (@Help) {
         chomp($HelpNumber);
         @TempHelp = split(/::/, $HelpNumber);
         if ($TempHelp[1] eq uc($params[1])) {
            if (&has_access($params[0]) >= $TempHelp[2]) {
               $HStop = "End::" . uc($params[1]);
               $HCount++;
               while (length($Help[$HCount]) ne 0) {
                  if ($HelpNumber =~ /^\#/) { # Its a comment ...
                     next;
                  }
                  chomp($Help[$HCount]);
                  if ($Help[$HCount] eq $HStop) {
                     return;
                  }
                  &notice($params[0],"$Help[$HCount]");
                  $HCount++;
               }
               return;
            } else {
               &notice($params[0],"You do not have access to view that help");
               return;
            }
         } 
         $HCount++;
      }
      &notice($params[0],"Command $params[1] does not exist");
      return;
   } else {
      $HCount = 0;
      foreach $ShowByLevel (@Help) {
         chomp($ShowByLevel);
         @TempHelp = split(/::/, $ShowByLevel);
         if ($TempHelp[1] == &has_access($params[0])) {
            $HStop == "End::" . &has_access($params[0]);
            $HCount++;
            while (length($Help[$HCount]) ne 0) {
               if ($ShowByLevel =~ /^\#/) {
                  next;
               }
               chomp($Help[$HCount]);
               if ($Help[$HCount] =~ /End::/) {
                  return;
               }
               &notice($params[0],"$Help[$HCount]");
               $HCount++;
            }
            return;
         }
         $HCount++;
      }
   }
   undef(@params);
}
###############
# Remote command SHUTDOWN
#
# Obviously this remote command shuts down
# the service. 
sub do_shutdown {
   $LastSub = "do_shutdown";
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) == 400) {
      $ShutDownNick = $params[0];
      splice(@params,0,1);
      $ShutDownReason = join(" ", @params);
      &globops("Shutdown command received by $ShutDownNick ($ShutDownReason) -- 6 second delay");
      &logit("$ShutDownNick shutdown FloodWorld -- $ShutDownReason -- 6 second delay");
      foreach $line (@Channels) {
         chomp($line);
         print $sock ":$GhostNick PART $line :.\n";
      }  
      $PauseTime = (time() + 6);
      while (time() <= $PauseTime) {
         $blah++; # Note needed but gives it something to do
      }
      print $sock ":$MyServerName QUIT :$ShutDownReason\n";
      &logdebug("--> :$MyServerName QUIT :$ShutDownReason") if (($DebugMode == 2) || ($DebugMode == 3));
      close DEBUG;
      exit;
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed SHUTDOWN (No access/ident)");
      &logit("$params[0] failed SHUTDOWN -- No access");
   }
   undef(@params);
}
###############
# Remote command STATUS
#
# This will first display any and
# all data stored to the database 
# from the TEXTFLOOD command and 
# what the individual settings are
# per line of text. Then it will
# display the amount of people that
# the service has removed from the 
# network and last but no least it
# will display the service uptime
sub do_status {
   $LastSub = "do_status";
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) >= 100) {
      $FCount = 0;
      &notice($params[0],"Current service status.");
      if ($ChanFloodTextHandle eq "off") {
         &notice($params[0],"Flood Text: off");
      } else {
         &notice($params[0],"Flood Text: on");
      }
      if ($CTCPFlood ne "off") {
         &notice($params[0],"CTCP Flood: $CTCPFlood CTCPs in $CTCPFloodTime Second(s)");
      } else {
         &notice($params[0],"CTCP Flood: off");
      }
      if ($NickFloodStatus eq "on") {
         &notice($params[0],"Nick Flood: $NickFloodCount changes in $NickFloodTime seconds");
      } else {
         &notice($params[0],"Nick Flood: off");
      }
      if ($misc{'AutoFlood'} eq "off") {
         &notice($params[0],"Auto Flood Detect: off");
      } else {
         &notice($params[0],"Auto Flood Detect: $AutoFloodCount lines in $AutoFloodTime seconds");
      }
      if ($GRNameStatus eq "off") {
         &notice($params[0],"GRName System: off");
      } else {
         &notice($params[0],"GRName System: on");
      }
      if ($JoinFloodStatus eq "on") {
         &notice($params[0],"Join Flood: $JoinFloodCount joins in $JoinFloodTime seconds");
      } else {
         &notice($params[0],"Join Flood: off");
      }
      if ($NoticeFloodStatus eq "on") {
         &notice($params[0],"Notice Flood: $NoticeFloodCount lines in $NoticeFloodTime seconds");
      } else {
         &notice($params[0],"Notice Flood: off");
      }
      &notice($params[0],"I currently have $TotalNicks users stored to memory");
      &notice($params[0],"Currently monitoring " . @Channels . " Channel(s)");
      &notice($params[0],"I have glined $TFGlines flooders since last restart");
      &notice($params[0],"Started: $LoadTime");
      &notice($params[0],"Now: " . time());
      &notice($params[0],"Up: " . &duration((time() - $LoadTime)));
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed status (No access/ident)");
   }
   undef(@params);
}
###############
# Remote command GRNAME
#
# This is perhaps one of the
# most useful parts of this
# service. Basically when you
# add a realname to the GRNAMES
# database it makes it so that 
# whenever anyone connects to the
# network that has that realname 
# will be banned. Note that this
# has to be exact, it is case
# sensitive and length sensitive.
# Any difference between one that
# was set and a user connecting with
# it, it will not work on that user.
sub do_grname {
   $LastSub = "do_grname";
   @params = split(/\s+/, shift);
   $RNKNick = $params[0];
   $What = $params[1];
   splice(@params,0,2);
   $RNFCount = 0;
   $FindRN = join(" ", @params);
   if (&has_access($RNKNick) >= 300) {
      if (lc($What) eq "add") {
         foreach $line (@GNames) {
            chomp($line);
            if (($line eq $FindRN) && ($line ne "")) {
               &notice($RNKNick,"That entry already exists.");
               return;
            }
         }
         push @GNames, $FindRN;
         &write_grnames; 
         &notice($RNKNick,"Added ($FindRN) to GLined Real Names database -- Scanning for matches");
         &globops("$RNKNick added ($FindRN) to GLined Real Names database -- Scanning for matches");
         if ($GRNameStatus eq "on") {
            foreach $RNFind (keys %nicks) {
               if ($nicks{$RNFind}->{RealName} eq $FindRN) {
                  if (($nicks{$RNFind}->{Oper} == 0) && (!(defined($ops{lc($RNFind)})))) {
                     $FoundOne = 1;
                     print $sock ":$MyClientName PRIVMSG $DebugChannel :GRNAME (COMMAND): $RNFind -- " . $nicks{lc($RNFind)}->{Host} . " -- REALNAME:($FindRN)\n" if ($UseDebug);
                     &loggline("GRNAME REALNAME -- $RNFind -- $FindRN");
                     &RemoveUser($sock,"*@" . $nicks{$RNFind}->{Host},"(FloodWorld) You have triggered a network automatic flood detection -- If you believe this is an error please contact $HelpHost immediately.");
                     &KillUser($sock,$RNFind,"(FloodWorld) Flood Detection Triggered") if ($UseKill);
                     $RNFCount++;
                     $TFGlines++;
                  }
               }
            }
         
            if ($FoundOne == 1) {
               &logit("$RNKNick did GRNAME on $FindRN");
               &globops("Matches found for ($FindRN) and removed -- Matched: $RNFCount");
               &notice($RNKNick,"Successfully removed all matching realnames");
            } else {
               &notice($RNKNick,"No Matching realname found.");
               &globops("No matches found on realname ($FindRN)");
            }
         } else {
            &globops("There isnt anything I can do, GRName is deactivated!");
         }
      } elsif (lc($What) eq "del") {
         if ($FindRN =~ /\D/) {
            &notice($RNKNick,"Syntax: GRNAME DEL Number");
            return;
         }
         if (!(defined($GNames[$FindRN]))) {
            &notice($RNKnick,"Entry $FindRN does not exist");
            return;
         }
         &logit("$RNKNick removed GLined Real Name ($GNames[$FindRN])");
         &notice($RNKNick,"Successully removed Real Name ($GNames[$FindRN]) from GLined Real Names database");
         &globops("$RNKNick removed GLined Real Name ($GNames[$FindRN]) from database");
         splice(@GNames,$FindRN,1);
         &write_grnames;
      } elsif (lc($What) eq "list") {
         $GRNCount = 0;
         &notice($RNKNick,"Listing all GLine Real Names in order from array");
         foreach $line (@GNames) {
            chomp($line);
            if ($line ne "") {
               &notice($RNKNick,"Number $GRNCount: $line");
            }
            $GRNCount++;
         }
      } elsif (lc($What) eq "off") {
         $misc{'GRName'} = "off";
         &write_misc;
         &globops("$RNKNick deactivated the GRName system");
         &logit("$RNKNick deactivated the GRName system");
      } elsif (lc($What) eq "on") {
         $misc{'GRName'} = "on";
         &write_misc;
         &globops("$RNKNick activated the GRName system -- Scanning for any matches on all GRNamed Realnames");
         foreach $GRUsers (keys %nicks) {
            foreach $GRNamesCheck (@GNames) {
               if ($GRNamesCheck eq $nicks{$GRUsers}->{RealName}) {
                  if (($nicks{$GRUsers}->{Oper} == 0) && (!(defined($ops{lc($GRUsers)})))) {
                     print $sock ":$MyClientName PRIVMSG $DebugChannel :GRNAME (COMMAND): $GRUsers -- " . $nicks{lc($GRUsers)}->{Host} . " -- REALNAME:($GRNamesCheck)\n" if ($UseDebug);
                     &loggline("GRNAME REALNAME -- $GRUsers -- $GRNamesCheck");
                     &RemoveUser($sock,"*@" . $nicks{$GRUsers}->{Host},"(FloodWorld) You have triggered a network automatic flood detection -- If you believe this is an error please contact $HelpHost immediately.");
                     &KillUser($sock,$GRUsers,"(FloodWorld) Flood Detection Triggered") if ($UseKill);
                     $RNFCount++;
                     $TFGlines++;
                  }
               }
            }
         }
         if ($RNFCount > 0) {
            &globops("Matches found -- Removed: $RNFCount");
         } else {
            &globops("No matches found for any realnames in the database");
         }
         $GRNameStatus = "on";           
      } else {
         &notice($RNKNick,"Syntax: GRNAME {ADD|DEL|LIST|ON|OFF} {REALNAME|NUMBER}");
         return;
      }
   } else {            
      &notice($RNKNick,"Access denied");
      &globops("$RNKNick failed grname (No access/ident)");
   }
   undef(@params);
}
###############
# Remote command CHLEVEL
#
# This is used to change the level of
# the specified operator to the new 
# specified level.
sub do_chlevel {
   $LastSub = "do_chlevel";
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) == 400) {
      if ($params[2] =~ /\D/) {
         &notice($params[0],"$params[2] is not a valid level");
         return;
      }
      if ($ops{lc($params[1])}->{Level} eq $params[2]) {
         &notice($params[0],"$params[1] is already that level");
         return;
      }
      if (!(defined($ops{lc($params[1])}))) {
         &notice($params[0],"User $params[1] not found in database");
         return;
      }
      $ops{lc($params[1])}->{Level} = $params[2];
      &notice($params[0],"Level for $params[1] changed to $params[2]");
      &globops("$params[0] changed level for $params[1] to $params[2]");
      &logit("$params[0] changed level for $params[1] to $params[2]");
      &write_ops;
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed chlevel (No access/ident)");
      &logit("$params[0] failed chlevel on $params[1] (Tried: $params[2])");
   }
   undef(@params);
}
###############
# Remote command REHASH
#
# This is split up into two sub
# routines. Essentially when triggered
# it will reload all service databases
# and the help file.
sub do_rehash {
   $LastSub = "do_rehash";
   if (&has_access($_[0]) == 400) {
      &globops("$_[0] is rehashing all service databases");
      &logit("$_[0] triggered rehash");
      &srehash(1);
   } else {
      &notice($_[0],"Access denied");
      &globops("$_[0] failed rehash (No access/ident)");
   }
   undef(@params);
}
###############
# Part two of the rehash.
#
# The reason it is split up into two sub 
# routines is so that the below sub can
# be used if this program receives SIGHUP
# from the operating system.
sub srehash {
   $LastSub = "do_rehash";
   if ($_[0] != 1) {
      &globops("Received SIGHUP -- Reloading all service databases and help file");
      &logit("Received SIGHUP -- Reinitializing all service databases");
   }
   &load_all;
   foreach $WipeGlines (keys %glined) {
      delete($glined{$WipeGlines});
   }
   undef(@params);
}
###############
# Remote command FORBID
#
# Used to add text to the database that
# is forbidden from being used as 
# TEXTFLOOD text. Good to use when 
# there are key words that are commonly
# said.
sub do_forbid {
   $LastSub = "do_forbid";
   @params = split(/\s+/, shift);
   if ((@params < 3) && ($params[1] ne "list")) {
      &notice($params[0],"Syntax: FORBID {ADD|DEL|LIST} {TEXT|NUMBER}");
      return;
   }
   $Who = $params[0];
   $What = $params[1];
   splice(@params,0,2);
   $Text = join(" ", @params);
   if (&has_access($Who) >= 300) {
      if ($What eq "add") {
         foreach $IsForb (@forbidden) {
            chomp($IsForb);
            if (lc($IsForb) eq lc($Text)) {
               &notice($Who,"That text already exists");
               return;
            }
         }
         push @forbidden, $Text . "\n";
         &notice($Who,"Added ($Text) to forbidden text");
         &globops("$Who added forbidden text ($Text) to database");
         &logit("$Who added text ($Text) to forbidden database");
         &write_forbidden;
      } elsif ($What eq "del") {
         if ($Text =~ /\D/) {
            &notice($Who,"Syntax: FORBID DEL Number");
            return;
         }
         if ($forbidden[$Text]) {
            chomp($forbidden[$Text]);
            &notice($Who,"Removed forbidden text ($forbidden[$Text]) from database");
            &globops("$Who removed forbidden text ($forbidden[$Text]) from database");
            &logit("$Who removed ($forbidden[$Text]) from forbidden database");
            splice(@forbidden,$Text,1);
            &write_forbidden;
            return;
         }
         &notice($Who,"Forbidden text number $Text does not exist");
         return;
      } elsif ($What eq "list") {
         $SFNum = 0;
         &notice($Who,"Listing forbidden text");
         foreach $ShowForb (@forbidden) {
            chomp($ShowForb);
            if ($ShowForb ne "") {
               &notice($Who,"Forbidden $SFNum: $ShowForb");
            }
            $SFNum++;
         }
      } else {
         &notice($Who,"Unknown option " . uc($Text));
         return;
      }
   } else {
       &notice($Who,"Access denied");
       &globops("$Who failed forbid (No access/ident)");
   }
   undef(@params);
}
###############
# Remote command INFO
# 
# This is not really a useful command
# but I added it in while debugging
# the service and I see no point in
# removing it.
sub do_info {
   $LastSub = "do_info";
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) >= 100) {
      if (defined($ops{lc($params[1])})) {
         if(defined($nicks{lc($params[1])})) { $IsOn = "Yes"; } else { $IsOn = "No"; }
         if($ops{lc($params[1])}->{Identified} eq 1) { $IsID = "Yes"; } else { $IsID = "No"; }
         &notice($params[0],"Information on $params[1]");
         &notice($params[0],"Is online: $IsOn");
         &notice($params[0],"Is identified: $IsID");
         &notice($params[0],"Level: " . $ops{lc($params[1])}->{Level});
         return;
      } else {
         &notice($params[0],"No such operator ($params[0])");
         return;
      }
   } else {
      &notice($params[0],"Access denied");
   }
   undef(@params);
}
###############
# Remote command CJOINALL
#
# This command is used to make the
# ghost client join all channels 
# from the database. Useful if it
# got kicked from some of the channels
# or anything like that. 
sub do_cjoinall {
   $LastChan = "do_cjoinall";
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) >= 100) {
      &client_join_chans;
      &notice($params[0],"Client has joined all channels in database.");
      &globops("$params[0] did cjoinall (Client Join All Channels)");
      &logit("$params[0] did cjoinall");
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed cjoinall (No access/ident)");
      &logit("$params[0] failed cjoinall");
   }
   undef(@params);
}
###############
# Remote command AFLOOD
#
# This was incorporated into the code to give 
# it the on-the-fly ability to change the number
# of lines and number of seconds on-the-fly.
# The syntax speaks for itself.
sub do_aflood {
   $LastSub = "do_aflood";
   @params = split(/\s+/, shift);
   if ($params[1] eq "") {
      &notice($params[0],"Syntax: AFLOOD {SET|OFF} NumberOfTime NumberOfSeconds");
      return;
   }
   if (&has_access($params[0]) >= 200) {
      if (lc($params[1]) eq "set") {
         if (($params[2] =~/\D/) || ($params[3] =~ /\D/)) {
            &notice($params[0],"Syntax: AFLOOD {SET|OFF} NumberOfTimes NumberOfSeconds");
            return;
         }
         $AutoFloodCount = $params[2];
         $AutoFloodTime = $params[3];
         $misc{'AutoFlood'} = $AutoFloodCount . "::" . $AutoFloodTime;
         &write_misc;
         &globops("$params[0] set Auto Flood Detect to $AutoFloodCount lines in $AutoFloodTime seconds");
         &notice($params[0],"Auto Flood Detect set to $AutoFloodCount lines in $AutoFloodTime seconds");
         &logit("$params[0] set Auto Flood Detect to $AutoFloodCount lines in $AutoFloodTime seconds");
      } elsif (lc($params[1]) eq "off") {
         $misc{'AutoFlood'} = "off";
         &write_misc;
         &notice($params[0],"Auto Flood Detection deactivated");
         &globops("$params[0] deactivated Auto Flood Detection");
         &logit("$params[0] deactivated Auto Flood Detection");
      } else {
         &notice($params[0],"Unknown parameter -- $params[1]");
      }   
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed AFLOOD (No access/ident)");
   }
   undef(@params);
}
###############
# Remote command NICKFLOOD
#
# This is used to either set the number of
# nick changes within the specified amount
# of seconds before nick flood is triggered,
# or it is used to deactivate the nick flood
# protection.
sub do_nickflood {
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) >= 200) {
      if (lc($params[1]) eq "set") {
         if (($params[2] =~ /\D/) || ($params[3] =~ /\D/)) {
            &notice($params[0],"Syntax: NICKFLOOD {SET|OFF} NumberOfTimes NumberOfSeconds");
            return;
         }
         if (($params[2] eq "") || ($params[3] eq "")) {
            &notice($params[0],"Syntax: NICKFLOOD {SET|OFF} NumberOfTimes NumberOfSeconds");
            return;
         }            
         $misc{'NickFlood'} = $params[2] . "::" . $params[3];
         &write_misc;
         &notice($params[0],"Nick Flood Protection set to $params[2] changes in $params[3] seconds");
         &globops("$params[0] set Nick Flood Protection to $params[2] changes in $params[3] seconds");
         &logit("$params[0] set Nick Flood Protection to $params[2] changes in $params[3] seconds");
         $NickFloodStatus = "on";
         $NickFloodCount = $params[2];
         $NickFloodTime = $params[3];
      } elsif (lc($params[1]) eq "off") {
         $misc{'NickFlood'} = "off";
         &write_misc;
         &notice($params[0],"Nick Flood Protection deactivated");
         &globops("$params[0] deactivated Nick Flood Protection");
         &logit("$params[0] deactivated Nick Flood Protection");
      } else {
         &notice($params[0],"Unknown parameter -- $params[1]");
      }
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed NICKFLOOD (No access/ident)");
   }
   undef(@params);
}
###############
# Remote command JOINFLOOD
#
# Basically this takes the parameters and finds out if we 
# are setting a join flood protection or not. If we are 
# then lets turn the system on and set it properly and
# store it to database. If it is deactivated then we
# will also note that in the database so if it is 
# restarted for one reason or another the protection will
# still be active. This is beneficial incase you have a 
# crontab on it and for some reason it loses connection
# and shuts itself down.
sub do_joinflood {
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) >= 200) {
      if (lc($params[1]) eq "set") {
         if ($params[2] < 4) {
            &notice($params[0],"Sorry, but you must specify atleast 4 joins");
            &globops("$params[0] failed JOINFLOOD (NumberOfJoins too low)");
            return;
         }
         if (($params[2] =~ /\D/) || ($params[3] =~ /\D/)) {
            &notice($params[0],"Syntax: JOINFLOOD {SET|OFF} NumberOfJoins NumberOfSeconds");
            return;
         }
         if ((!($params[2])) || (!($params[3]))) {
            &notice($params[0],"Syntax: JOINFLOOD {SET|OFF} NumberOfJoins NumberOfSeconds");
            return;
         }
         $misc{'JoinFlood'} = $params[2] . "::" . $params[3];
         &write_misc;
         $JoinFloodCount = $params[2];
         $JoinFloodTime = $params[3];
         $JoinFloodStatus = "on";
         &logit("$params[0] set JOINFLOOD -- $JoinFloodCount joins in $JoinFloodTime seconds");
         &notice($params[0],"Join Flood protection set to $JoinFloodCount joins in $JoinFloodTime seconds");
         &globops("$params[0] set Join Flood protection to $JoinFloodCount joins in $JoinFloodTime seconds");
      } elsif (lc($params[1]) eq "off") {
         $misc{'JoinFlood'} = "off";
         &write_misc;
         $JoinFloodStatus = "off";
         &logit("$params[0] deactivated JOINFLOOD");
         &notice($params[0],"Join Flood protection deactivated");
         &globops("$params[0] deactivated Join Flood protection");
      } else {
         &notice($params[0],"Unknown parameter -- $params[1]");
      }
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed JOINFLOOD (No access/ident)");
   }
}
###############
# Remote command GHOSTHOST
#
# This will log off the specified ghost
# client and introduce a new one with the
# specified nickname and host. 
#
# Special thanks to RioTeer for coming up
# with this idea and giving me permission
# to put it in the public release.
sub do_ghosthost {
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) >= 200) {
      if ((!($params[1])) || (!($params[2])) || (!($params[3]))) {
         &notice($params[0],"Syntax: GHOSTHOST CurrentGhostNick NewNick NewHost");
         return;
      }
      print $sock ":$params[1] QUIT :gone\n";
      &logdebug("--> :$params[1] QUIT :gone") if (($DebugMode == 2) || ($DebugMode == 3));
      &IntroduceUser($sock,$params[2],$params[3]);      
      foreach $JChan (@Channels) {
         &SendJoinChan($sock,$JChan,$params[2]);
      }
      &notice($params[0],"GHOSTHOST Successful");
      &globops("$params[0] did GHOSTHOST -- Old($params[1]) -- New($params[2])");
      &logit("$params[0] did GHOSTHOST -- Old Ghost($params[1]) -- New Ghost($params[2])");
      $GhostNick = $params[2];
   } else {
      &logit("$params[0] failed GHOSTHOST -- No Access/Ident");
      &notice($params[0],"Access Denied");
      &globops("$params[0] failed GHOSTHOST (No access/ident)");
   }
}
###############
# NFLOOD remote command
#
# Set the settings for notice flood protection
sub do_nflood {
   @params = split(/\s+/, shift);
   if (&has_access($params[0]) >= 200) {
      if (lc($params[1]) eq "set") {
         if (($params[2] =~ /\D/) || ($params[3] =~ /\D/)) {
            &notice($params[0],"Syntax: NFLOOD SET NumberOfTimes NumberOfSeconds");
            return;
         }
         if ($params[2] < 4) {
            &notice($params[0],"ERROR: You must specify atleast 4 lines");
            &globops("$params[0] attempted to set Auto-Notice Flood detection to less than 4 lines");
            return;
         }
         $NoticeFloodStatus = "on";
         $NoticeFloodCount = $params[2];
         $NoticeFloodTime = $params[3];
         $misc{'NoticeFlood'} = $NoticeFloodCount . "::" . $NoticeFloodTime;
         &write_misc;
         &globops("$params[0] set Auto-Notice Flood detection to $params[2] lines in $params[3] seconds");
         &notice($params[0],"Auto-Notice Flood detection set to $params[2] lines in $params[3] seconds");
         &logit("$params[0] set Auto-Notice Flood detection to $params[2] lines in $params[3] seconds");
      } elsif ($params[1] eq "off") {
         $NoticeFloodStatus = "off";
         $NoticeFloodCount = 100;
         $NoticeFloodTime = 100;
         $misc{'NoticeFlood'} = "off";
         &write_misc;
         &globops("$params[0] deactivated Auto-Notice Flood detection");
         &notice($params[0],"Auto-Notice Flood detection deactivated");
         &logit("$params[0] deactived Auto-Notice Flood detection");
      } else {
         &notice($params[0],"Unknown parameter -- $params[1]");
      }
   } else {
      &notice($params[0],"Access denied");
      &globops("$params[0] failed NFLOOD (No access/ident)");
      &logit("$params[0] failed NFLOOD -- No access/ident)");
   }
}
sub set_flood_settings {
   ###############
   # Lets get the data needed for the auto-flood detection
   # system.
   if ($misc{'AutoFlood'} ne "off") {
      @GetAutoData = split(/::/,$misc{'AutoFlood'});
      $AutoFloodCount = $GetAutoData[0];
      $AutoFloodTime = $GetAutoData[1];
      undef(@GetAutoData);
   }

   ###############
   # Now lets see if textflood should be active
   if ($misc{'TextFlood'} ne "off") {
      $ChanFloodTextHandle = "on";
   }

   ###############
   # Lets see if we should have CTCP Flood protection
   # active.
   if ($misc{'CTCPFlood'} ne "off") {
      @TempC = split(/::/,$misc{'CTCPFlood'});
      $CTCPFlood = $TempC[0];
      $CTCPFloodTime = $TempC[1];
      undef(@TempC);
   } else {
      $CTCPFlood = "off";
   }

   ###############
   # Lets see if nick flood protection is off
   # or has settings
   if ($misc{'NickFlood'} eq "off") {
      $NickFloodStatus = "off";
   } else {
      $NickFloodStatus = "on";
      @TempN = split(/::/,$misc{'NickFlood'});
      $NickFloodCount = $TempN[0];
      $NickFloodTime = $TempN[1];
      undef(@TempN);
   }

   ###############
   # Now lets see if join flood protection is off
   # or has settings
   if ($misc{'JoinFlood'} eq "off") {
      $JoinFloodStatus = "off";
      ###############
      # The following two scalars are
      # here as a failsafe. 
      $JoinFloodCount = 100;
      $JoinFloodTime = 1; 
   } else {
      $JoinFloodStatus = "on";
      @TempN = split(/::/,$misc{'JoinFlood'});
      $JoinFloodCount = $TempN[0];
      $JoinFloodTime = $TempN[1];
      undef(@TempN);
   }

   ###############
   # Now lets see if the GRName detection system
   # is on or off. 
   if ($misc{'GRName'} eq "off") {
      $GRNameStatus = "off";
   } else {
      $GRNameStatus = "on";
   }

   ###############
   # Notice flood settings retrieved
   # from misc.db
   if ($misc{'NoticeFlood'} eq "off") {
      $NoticeFloodStatus = "off";
      # Failsafes
      $NoticeFloodCount = 100;
      $NoticeFloodTime = 100;
   } else {
      $NoticeFloodStatus = "on";
      @TempN = split(/::/, $misc{'NoticeFlood'});
      $NoticeFloodCount = $TempN[0];
      $NoticeFloodTime = $TempN[1];
   }
}
sub encrypt {
   $EPass = shift;
   $E_TempPass = `md5/fwmd5 $EPass`;
   chomp($E_TempPass);
   return $E_TempPass;
}
