#!/usr/bin/perl
#
# FloodWorld v1.0.a3 (Protocol Definition Perl Module)
#
# (C)Copyright 2001
# Brian M. King
# All Rights Reserved
#
# This program is provided as is, without warranty. The author is not liable
# for any and all actions as a result of the code herein.
#
# Special thanks to PreZ for helping me with desiging this
# protocol system. Without his help I dont think I would have
# gotten it done this fast. :)
#
# Also, special thanks to Shrapnal for helping me with beta
# testing and debugging the software as well as providing 
# the criticism and ideas that helps to make this the best
# software there is =)
#
# I have developed a way to have a certain string placed into
# the protocol system with certain flags in it. What this does
# is it enables me to use a single function to use that string
# no matter what protocol its defined in to determine what 
# needs to be passed to the remote server and how exactly it
# needs to be passed to it. Here is a list of the supported
# flags that I have implemented into the system.
#
# <NN> = NickName
# <TM> = Time
# <UM> = User Modes
# <CH> = Client Host
# <VH> = Vanity Host
# <SN> = Server Name
# <SD> = Server Description
# <RN> = Real Name
# <LP> = Link Pass
# <WHO> = NickName (For JoinChan protocol setting)
# <CHAN> = #Channel (For JoinChan protocol setting)
# <RU> = Remove User (The host of the user to be removed)
# <R> = Reason for removing that user
#
# Now, one might think "Why have a different section 
# for each ircd". Well, the reason for that is that
# each setting has to be exact and if one thing is
# different between each ircd then the entire protocol
# will mess up and so in turn the service will not work
# at all. That is why you will see that most of the 
# things between each ircd is the same. Now, let me 
# explain each item in the protocol system so that if
# you want to add another protocol to this service, you
# will know how.
#
# As you can see the entire protocol is based on the
# protocol hash. 
#
# NickString Meanings ...
# Command  -- 1 means NICK is the first word, 2 means its the second
# NickName -- Array element 0 (Parsed using space as a delimiter from
#             the NICK string. Remember, it is 0 because as a standard
#             all arrays start with the first element as 0.)
# UModes   -- If UserModes are defined in the NICK string then this is 
#             the token number in which to read from the array.
# Ident    -- Token number for the users Ident from the array
# Host     -- Real Host token number
# RealName -- This is the starting point of this realname token. Remember
#             that RealNames can be multiple words. This is the token 
#             number for the first word of the realname.
#
# Other ...
# IntroduceUser -- This is the NICK string that will be sent to the 
#                  remote server to log the main client (FloodWorld)
#                  and the ghost client onto the network. This uses
#                  the above mentioned flags to accomplish this.
#
# HasGlobOps -- This means that network supports the /GLOBOPS oper
#               command. 1 = Yes, 0 = No
#
# ServerString -- This is the raw server string syntax that is sent
#                 to the remote server for linking. 
#
# LinkPass -- Obviously this is the linkpass string sent to the remote
#             server on connect.
#
# NeedProtoString -- Some IRCd's do not define the protocol to use when
#                    you link to them in the SERVER string. If not then
#                    we will just send them a protocol string to let 
#                    them know which protocol we want to use.
#
# Proto -- This is the protocol string that defines which protocol we 
#          want to use when we link. Coinsides with the above mentioned
#          protocol object.
#
# JoinChan -- This is the string in which the service will use in order to
#             make the ghost client join the supported channels.
#
# NeedChop -- Contrary to popular belief not all IRCd's tack on a carriage
#             return AND line feed to the end of a string before they send
#             it through to us. So if you set this value to 1 then we will
#             chop off the last charecter in the string after we strip the
#             carriage return off of it. If its 0 then we will just remove
#             the carriage return.
#
# EndOfBurst -- This means the IRCd we're linking too has some sort of
#               string that signifies that we have reach the end of burst
#               and so send this message through.
#
# ExpectEOB -- The remote server sends this as a $Command to signify that
#              it has finished sending me all the data I need.
#
# NeedUModes -- This means that UserModes are not specified in the NICK
#               string so we need to send in our clients usermodes 
#               seperately.
#
# JoinTrigger -- This word is what the service will use to determine what the
#                syntax word is to specify that the ircd is introducing a user
#                to that channel. Some IRCd's use SJOIN while others use just
#                JOIN. If you add another protocol to this service you must
#                specify this or the JoinFlood Protection will not work. You
#                must remember that anything proceding this protocol object
#                will start as token number 0. Each work from there on will
#                be incremented by 1. See examples below.
#
# JoinNick -- This is the token number of where to find the users nickname in 
#             the join string if it procedes the JoinTrigger. If it precedes
#             it then define this as "source". That indicates to the service
#             that we're looking for the users nickname as the first word
#             in the JoinTrigger string.
#
# JoinChannel -- This is the token number of the channel specification in the
#                JoinTrigger string. Please view the below examples.
#
# Example Join String 1:
#    :my.server.net SJOIN 0 0 #SomeChannel + :Raider
#
#    JoinTrigger = SJOIN
#    JoinNick = 4
#    JoinChannel = 2
#
# Example Join String 2:
#    :Raider JOIN :#SomeChannel
#
#    JoinTrigger = JOIN
#    JoinNick = source
#    JoinChannel = 0
#
# Example Nick String 1:
#    NICK Raider 1 9000000000 +oaA Raider 123.rochester.rr.com my.server.net 0 :My Realname
#
#    NICK = 1
#    NickName = 0
#    Umodes = 3
#    Ident = 4
#    Host = 5
#    RealName = 8
#
#    IntroduceUser: NICK <NN> 1 <TM> +<UM> <NN> <CH> <SN> 0 :<RN>
#
# Example Nick String 2:
#    :my.server.net NICK Raider 1 9000000000 +oaA Raider 123.rochester.rr.com 0 :My Realname
#
#    NICK = 2
#    NickName = 0
#    Umodes = 3
#    Ident = 4
#    Host = 5
#    RealName = 7
#
#    IntroduceUser: :<SN> NICK <NN> 1 <TM> +<UM> <NN> <CH> 0 :<RN>
#
#
# NOTE: If you change ANYTHING in this file you will need to restart the
#       service before it has any effect.
#
sub set_protocol_standards {
   ###############
   # Relic IRCd v4.x preX
   if ($config{'Protocol'} == 1) {
      $protocol{'NickString'}->{Command}  = 1;
      $protocol{'NickString'}->{NickName} = 0;
      $protocol{'NickString'}->{UModes}   = 3;
      $protocol{'NickString'}->{Ident}    = 4;
      $protocol{'NickString'}->{Host}     = 5;
      $protocol{'NickString'}->{RealName} = 9;
 
      $protocol{'IntroduceUser'}          = "NICK <NN> 1 <TM> +<UM> <NN> <CH> <VH> <SN> 0 :<RN>";
      
      $protocol{'HasGlobOps'}             = 1;

      $protocol{'ServerString'}           = "SERVER <SN> 1 :<SD>";
      $protocol{'LinkPass'}               = "PASS <LP> :TS";

      $protocol{'NeedProtoString'}        = 1;
      $protocol{'Proto'}                  = "CAPAB TS3";

      $protocol{'JoinChan'}               = "SJOIN 2 2 <CHAN> + :<WHO> ";

      $protocol{'NeedChop'}               = 1;

      $protocol{'EndOfBurst'}             = ":$MyServerName EOB";

      $protocol{'ExpectEOB'}              = "EOB";

      $protocol{'JoinTrigger'}            = "SJOIN";
      $protocol{'JoinNick'}               = 4;
      $protocol{'JoinChannel'}            = 2;

   } elsif ($config{'Protocol'} == 2) {
      ###############
      # DreamForge 4.6.7 
      $protocol{'NickString'}->{Command}  = 1;
      $protocol{'NickString'}->{NickName} = 0;
      $protocol{'NickString'}->{Ident}    = 3;
      $protocol{'NickString'}->{Host}     = 4;
      $protocol{'NickString'}->{RealName} = 7;

      $protocol{'IntroduceUser'}          = "NICK <NN> 1 <TM> <NN> <CH> <SN> 0 :<RN>";

      $protocol{'HasGlobOps'}             = 1;

      $protocol{'ServerString'}           = "SERVER <SN> 1 <TM> <TM> J11 :<SD>";
      $protocol{'LinkPass'}               = "PASS :<LP>";

      $protocol{'NeedProtoString'}        = 0;

      $protocol{'JoinChan'}               = ":<WHO> JOIN <CHAN>";

      $protocol{'NeedChop'}               = 1;

      $protocol{'NeedUModes'}             = 1;

      $protocol{'JoinTrigger'}            = "JOIN";
      $protocol{'JoinNick'}               = "source";
      $protocol{'JoinChannel'}            = 0;

   } elsif ($config{'Protocol'} == 3) {
      ###############
      # Bahamut 1.4.7
      $protocol{'NickString'}->{Command}  = 1;
      $protocol{'NickString'}->{NickName} = 0;
      $protocol{'NickString'}->{UModes}   = 3;
      $protocol{'NickString'}->{Ident}    = 4;
      $protocol{'NickString'}->{Host}     = 5;
      $protocol{'NickString'}->{RealName} = 8;

      $protocol{'IntroduceUser'}          = "NICK <NN> 1 <TM> +<UM> <NN> <CH> <SN> 0 :<RN>";

      $protocol{'HasGlobOps'}             = 1;

      $protocol{'ServerString'}           = "SERVER <SN> 1 :<SD>";
      $protocol{'LinkPass'}               = "PASS <LP> :TS";

      $protocol{'NeedProtoString'}        = 1;
      $protocol{'Proto'}                  = "CAPAB TS3";

      $protocol{'JoinChan'}               = ":<SN> SJOIN <TM> <TM> <CHAN> + :<WHO> ";

      $protocol{'NeedChop'}               = 1;

      $protocol{'JoinTrigger'}            = "SJOIN";
      $protocol{'JoinNick'}               = 4;
      $protocol{'JoinChannel'}            = 2;

   } elsif ($config{'Protocol'} == 4) {
      ###############
      # Bahamut 1.4.21
      $protocol{'NickString'}->{Command}  = 1;
      $protocol{'NickString'}->{NickName} = 0;
      $protocol{'NickString'}->{UModes}   = 3;
      $protocol{'NickString'}->{Ident}    = 4;
      $protocol{'NickString'}->{Host}     = 5;
      $protocol{'NickString'}->{RealName} = 8;

      $protocol{'IntroduceUser'}          = "NICK <NN> 1 <TM> +<UM> <NN> <CH> <SN> 0 :<RN>";

      $protocol{'HasGlobOps'}             = 1;

      $protocol{'ServerString'}           = "SERVER <SN> 1 :<SD>";
      $protocol{'LinkPass'}               = "PASS <LP> :TS";

      $protocol{'NeedProtoString'}        = 1;
      $protocol{'Proto'}                  = "CAPAB TS3 NOQUIT SSJOIN BURST UNCONNECT ZIP";

      $protocol{'JoinChan'}               = ":<SN> SJOIN <TM> <TM> <CHAN> + :<WHO> ";
      
      $protocol{'NeedChop'}               = 0;

      $protocol{'JoinTrigger'}            = "SJOIN";
      $protocol{'JoinNick'}               = 4;
      $protocol{'JoinChannel'}            = 2;

   } elsif ($config{'Protocol'} == 5) {
      ###############
      # Hybrid 6.0
      $protocol{'NickString'}->{Command}  = 1;
      $protocol{'NickString'}->{NickName} = 0;
      $protocol{'NickString'}->{UModes}   = 3;
      $protocol{'NickString'}->{Ident}    = 4;
      $protocol{'NickString'}->{Host}     = 5;
      $protocol{'NickString'}->{RealName} = 7;

      $protocol{'IntroduceUser'}          = "NICK <NN> 1 <TM> +<UM> <NN> <CH> <SN> :<RN>";

      $protocol{'HasGlobOps'}             = 0;

      $protocol{'ServerString'}           = "SERVER <SN> 1 :<SD>";
      $protocol{'LinkPass'}               = "PASS <LP> :TS";

      $protocol{'NeedProtoString'}        = 1;
      $protocol{'Proto'}                  = "CAPAB :QS EX CHW DE";

      $protocol{'JoinChan'}               = ":<SN> SJOIN <TM> <CHAN> + :<WHO> ";
      
      $protocol{'NeedChop'}               = 1;

      $protocol{'JoinTrigger'}            = "SJOIN";
      $protocol{'JoinNick'}               = 3;
      $protocol{'JoinChannel'}            = 1;

   } elsif ($config{'Protocol'} == 6) {
      ###############
      # Unreal 3.1
      $protocol{'NickString'}->{Command}  = 1;
      $protocol{'NickString'}->{NickName} = 0;
      $protocol{'NickString'}->{Ident}    = 3;
      $protocol{'NickString'}->{Host}     = 4;
      $protocol{'NickString'}->{RealName} = 7;

      $protocol{'IntroduceUser'}          = "NICK <NN> 1 <TM> <NN> <CH> <SN> 0 :<RN>";

      $protocol{'HasGlobOps'}             = 1;

      $protocol{'ServerString'}           = "SERVER <SN> 1 :<SD>";
      $protocol{'LinkPass'}               = "PASS :<LP>";

      $protocol{'NeedProtoString'}        = 1;
      $protocol{'Proto'}                  = "PROTOCTL NOQUIT SJOIN UMODE2";

      $protocol{'JoinChan'}               = ":<WHO> JOIN <CHAN>";

      $protocol{'NeedChop'}               = 1;

      $protocol{'NeedUModes'}             = 1;

      $protocol{'JoinTrigger'}            = "JOIN";
      $protocol{'JoinNick'}               = "source";
      $protocol{'JoinChannel'}            = 0;

   }
}
############
# Load another client onto the network
# with the specified nickname and the
# specified host. Originally created
# for the GHOSTHOST command but can
# be used elsewhere if needed.
sub IntroduceUser {
   $SSock = shift;
   $NewNick = shift;
   $NewHost = shift;
   $TimeStamp = time();
   $SString = $protocol{'IntroduceUser'};
   $SString =~ s/<NN>/$NewNick/g;
   $SString =~ s/<CH>/$NewHost/g;
   $SString =~ s/<TM>/$TimeStamp/g;
   $SString =~ s/<VH>/$NewHost/g;
   $SString =~ s/<SN>/$MyServerName/g;
   $SString =~ s/<RN>/$NewNick/g;
   $SString =~ s/<UM>/$GhostUmodes/g;
   print $sock "$SString\n";
   if (($DebugMode == 2) || ($Debug == 1)) {
      &logdebug("--> $SString");
   }
   undef($SSock);
   undef($SString);
}
###############
# This is used in conjunction with the
# KillUser protocol string. Only used
# if $UsedKill is defined as 1 in the
# config.pm and is only send to the server
# AFTER the gline/akill has been sent
# through (In case your oper service is
# lagging a bit).
sub KillUser {
   $SSock = shift;
   $Who = shift;
   $Reason = shift;
   $Path = "$MyServerName!$MyClientHost!$MyClientName";
   $SString = $protocol{'KillUser'};
   $SString =~ s/<NN>/$MyClientName/g;
   $SString =~ s/<CN>/$Who/g;
   $SString =~ s/<PATH>/$Path/g;
   $SString =~ s/<R>/$Reason/g;
   print $SSock "$SString\n" if (!($NoBan));
   if (($DebugMode == 2) || ($Debug == 1)) {
      &logdebug("--> $SString");
   }      
   undef($SSock);
   undef($SString);
   return;
}
###############
# This routine is what we will use to remove the
# user from the network. Note that this only 
# supports the flags <NN>, <RU>, and <R>, anything
# else will be ignored. 
sub RemoveUser {
   $SSock = shift;
   $RemoveHost = shift;
   $Reason = shift;
   $GT = length($RemoveHost); 
   $SString = $protocol{'RemoveUser'};
   $SString =~ s/<NN>/$MyClientName/g;
   $SString =~ s/<RU>/$RemoveHost/g;
   $SString =~ s/<R>/$Reason/g;
   $SString =~ s/<SN>/$MyServerName/g;
   $SString =~ s/<GT>/$GT/g;
   print $SSock $SString . "\n" if (!($NoBan));
   if (($DebugMode == 2) || ($Debug == 1)) {
      &logdebug("--> $SString");
   }
   undef($SSock);
   undef($SString);
}
###############
# This is used when we make a connection to
# the remote server. Once we connect to it 
# we need to send out link pass so what we
# will do is we will find out what the format
# is by accessing the protocol. Then we will
# make sure it is the correct format and send
# it through.
sub SendLinkPass {
   $SSock = shift;
   $TimeStamp = time();
   $SString = $protocol{'LinkPass'};
   $SString =~ s/<LP>/$MyLinkPass/g;
   $SString =~ s/<TM>/$TimeStamp/g;
   print $SSock $SString . "\n";
   if (($DebugMode == 2) || ($Debug == 1)) {
      &logdebug("--> $SString");
   }
   undef($SSock);
   undef($SString);
}
###############
# When we want to make the client join the
# specified channels we need to know what 
# format to use so lets find out by asking
# the protocol what to do. Once we know we 
# will then send it through to the remote
# server.
#
# Thanks to Duke for finding a bug in here
# and pointing it out.
sub SendJoinChan {
   $SSock = shift;
   $WhatChan = shift;
   $Who = shift;
   $TimeStamp = time();
   $SString = $protocol{'JoinChan'};
   $SString =~ s/<CHAN>/$WhatChan/g;
   $SString =~ s/<WHO>/$Who/g;
   $SString =~ s/<TM>/$TimeStamp/g;
   print $SSock $SString . "\n";
   if (($DebugMode == 2) || ($Debug == 1)) {
      &logdebug("--> $SString");
   }
   undef($SSock);
   undef($SString);
}
###############
# Ok, as above we need to know how the SERVER
# string should be formatted so that the remote
# server will let us link. So, lets ask the 
# protocol again then send out the proper format.
sub SendServerString {
   $SSock = shift;
   $TimeStamp = time();
   $SString = $protocol{'ServerString'};
   $SString =~ s/<SN>/$MyServerName/g;
   $SString =~ s/<SD>/$MyServerDescription/g;
   $SString =~ s/<TM>/$TimeStamp/g;
   print $SSock $SString . "\n";
   if (($DebugMode == 2) || ($Debug == 1)) {
      &logdebug("--> $SString");
   }
   undef($SSock);
   undef($SString);
}
###############
# This is by far the biggest headache I ever got
# while making this. What I did was, as above, 
# make this get the raw string from the protocol
# then format it to the proper way and then
# send it through to the remote server. As I mentioned
# some IRCd's have the same NICK string format while
# others do not but since there was another difference
# in the protocol I have to make an entire profile for
# it.
sub SignOnUser {
   $SSock = shift;
   $Which = shift;
   $TimeStamp = time();
   $SString = $protocol{'IntroduceUser'};
   if ($Which == 1) {
      $SString =~ s/<NN>/$MyClientName/g;
      $SString =~ s/<UM>/$MyClientUmodes/g;
      $SString =~ s/<CH>/$MyClientHost/g;
      $SString =~ s/<VH>/$MyClientVHost/g;
      $SString =~ s/<RN>/$MyClientDescription/g;
   } elsif ($Which == 2) {
      $SString =~ s/<NN>/$GhostNick/g;
      $SString =~ s/<UM>/$GhostUmodes/g;
      $SString =~ s/<CH>/$GhostHost/g;
      $SString =~ s/<VH>/$GhostVHost/g;
      $SString =~ s/<RN>/$GhostDescription/g;
   }
   $SString =~ s/<SN>/$MyServerName/g;
   $SString =~ s/<TM>/$TimeStamp/;
   print $SSock $SString . "\n";
   if (($Which == 1) && ($protocol{'NeedUModes'})) {
      print $SSock ":$MyClientName MODE $MyClientName :+" . $MyClientUmodes . "\n";
   } elsif (($Which == 2) && ($protocol{'NeedUModes'})) {
      print $SSock ":$GhostNick MODE $GhostNick :+" . $GhostUmodes . "\n";
   }
   if (($DebugMode == 2) || ($Debug == 1)) {
      &logdebug("--> $SString");
   }
   undef($SSock);
   undef($SString);
}
# DO NOT REMOVE THE BELOW LINE OTHERWISE I WONT WORK !!!
1
